#!/usr/bin/perl
## Version control: this was created 6/1/2021 by David Ovens
##  on lightning as getgfs0p25_and_check_with_wgrib2.pl.forexport
##  and tarred and sent to Ryan Lyman with this:
##  tar cvf ovensgetgfs0p25.tar getgfs0p25_and_check_with_wgrib2.pl wgrib2
##
## by David Ovens, 4/29/2021
## to be run as a cron job on lightning by ldm with this entry:
# format:
# min    hour  d-o-m month d-o-w command
# 0-59   0-23  1-31  1-12  0-6
#                          0=sun
######################################################################
## run_namaltinit scripts.
#10 8,20 * * * /usr/local/ldm/local/bin/notDST "/home/disk/mm5rt/bin/getgfs0p25_and_check_with_wgrib2.pl > /home/lightning2/tmp/input.grids/getgfs0p25grids.log 2>&1"
#10 9,21 * * * /usr/local/ldm/local/bin/isDST "/home/disk/mm5rt/bin/getgfs0p25_and_check_with_wgrib2.pl > /home/lightning2/tmp/input.grids/getgfs0p25grids.log 2>&1"

$res = "0p25";
if ($0 =~ /gfs0p5/) {
    $res = "0p50";
} elsif ($0 =~ /gfs1p00/) {
    $res = "1p00";
}
$default_ymdh = &time_increment(time,"i4",0,-12200,"YYYYMMDDH12");

$usage = "$0
  Retrieves GFS 0.25-degree grids from possible sites.

Usage; $0 [options]
  where options are:
   YYYYMMDDHH       --> use this time instead of the DEFAULT
   -debug[=N]       --> verbose output [optional level N=1,2,3]
   -help            --> print this help message
-hours x,y,..z    || x,y,...z ||
-hours h1-hn-hint   --> hours h1 to hn by hint, Default does hours 0-180-3
  -skip other       --> where other = sage, ftpprd, nomads
                        do not retrieve from this source";


$force = 0;
$timelimit = 240;
$onlymissing = 1; # skip copying conduit files if they're full size
$sleep = 240; # wait 4 minutes between retries
chomp($thismachine = `uname -n`);
$sagemainrundir = "/home/disk/sage1/mm5rt/data/input.grids";
$sageip     = "140.142.30.14";
$freshairip = "140.142.30.21";
$mainconduit = $mainsageconduit = "/home/disk/sage1/conduit";
if ($thismachine eq "frosty") {
    $mainrundir = "/home/disk/frosty/ovens/nobackup/input.grids";
    @checkothers = ("ftpprd", "nomads"); # will add "sage" if $ymdh < $default_ymdh
    $mainconduit = "/home/disk/data/conduit";
    $nonsage = 1;
} elsif ($thismachine eq "lightning") {
    # this is the normal location
    $mainrundir = "/home/lightning2/tmp/input.grids";
    @checkothers = ("nomads", "ftpprd"); # will add "sage" if $ymdh < $default_ymdh
    $mainconduit = "/home/disk/data/conduit";
    $nonsage = 1;
} elsif ($thismachine eq "sage") {
    $mainrundir = $sagemainrundir;
}

&load_wgrib2_protos();
$skip{sage} = 1; # default for the export version

if ("$ENV{LOGNAME}" ne "mm5rt" && "$ENV{LOGNAME}" ne "ovens") {
   print "this is only set up to run as user mm5rt\n";
   exit 1;
}
for ($i = 0; $i <= $#ARGV; $i++) {
    if ($ARGV[$i] =~ /^-ho/) {
	$i++;
	if ($ARGV[$i] =~ /^(\d+)-(\d+)(-(\d+))?$/) {
	    $int = 1;
	    $int = $4 if ($4 ne "");
	    for ($j = $1; $j <= $2; $j += $int) {
		push @hours,sprintf "%3.3d",$j;
	    }
	} else {
	    @hours = split(',',$ARGV[$i]);
	    for ($j = 0; $j <= $#hours; $j++) {
		$hours[$j] = sprintf "%3.3d",$hours[$j];
	    }
	}
    } elsif ("$ARGV[$i]" =~ /^-h/) {
	print "$usage\n";
	exit;
    } elsif ("$ARGV[$i]" =~ /^-f/) {
	$force = 1;
    } elsif ("$ARGV[$i]" =~ /^-sk/) {
	$skip{$ARGV[++$i]} = 1;
    } elsif ("$ARGV[$i]" eq "") {
	$ymdh = &time_increment(time,"i4",0,-12200,"YYYYMMDDH12");
    } else {
	$ymdh = $ARGV[$i];
    }
}
unless ($ymdh) {
    $ymdh = $default_ymdh;
}
if ($ymdh < $default_ymdh) {
    unshift @checkothers,"sage";
}
$ymd = substr($ymdh,0,8);
$yymmdd = substr($ymdh,2,6);
$hh  = substr($ymdh,-2);
$rundir = "$mainrundir/$ymdh";
system "mkdir -p $rundir" if (! -d "$rundir");
chdir "$rundir";
if ($nonsage) {
    $com = "ssh mm5rt\@$sageip mkdir -p $sagemainrundir/${ymdh}";
} else {
    $com = "mkdir -p $sagemainrundir/${ymdh}";
}
system "$com";

$dir{conduit}       = "$mainconduit/gfs.$ymd/$hh/atmos";
$dir{sageconduit}   = "$mainsageconduit/gfs.$ymd/$hh/atmos";
$dir{conduitstatus} = "$mainconduit/status/gfs.$ymd/$hh/atmos";
$endhr{"00"} = $endhr{"12"} = 180;
$endhr{"06"} = $endhr{"18"} = 72;

unless (@hours) {
    for ($j = 0; $j <= $endhr{"$hh"}; $j += 3) {
	push @hours,sprintf "%3.3d",$j;
    }
}

$retries = 0;
RETRY:
foreach $ihr (@hours) {
    $gfsfile   = "gfs.t${hh}z.pgrb2.$res.f$ihr";
    $localfile = "$rundir/$gfsfile";
    $sagefile  = "$sagemainrundir/${ymdh}/$gfsfile";
    $localstatusfile = "$rundir/$gfsfile.incoming";
    if (-e "$localfile") {
	&scan_wgrib2("$localfile",$ihr);
    } else {
	$missing = "$localfile";
    }
    if ($missing eq "") {
	if ($localfile eq $sagefile) {
	    next;
	} else {
	    ##
	    ## our local version is good, let's rsync it to sage
	    ## even if domain name service is down
	    $com = "rsync -av $localfile mm5rt\@$sageip:$sagefile";
	    print "$com\n";
	    $iret = system "$com";
	    print "iret = $iret\n";
	    if ($ENV{LOGNAME} =~ /mm5rt/ || $ENV{LOGNAME} eq "ldm") {
		$com = "rsync -av $localfile ldm\@$sageip:$dir{sageconduit}";
		print "$com\n";
		$iret = system "$com";
		print "iret = $iret\n";
	    }
	    next;
	}
    }
    print "missing $missing\n";
    $fstat = 1;
    if (-e "$dir{conduit}/$gfsfile") {
	$iret = &check_conduit_status($gfsfile);
	if ($iret == 0) {
	    $conduitminsize  = $expected{$file}{this};
	    $conduitfullsize = $expected{$file}{full};
	} elsif ($iret == 1) {
	    # file is still arriving
	    $itry = 0;
	    while ($itry < 4 && $iret == 1) {
		$now = &time_increment(time,"i4",0,0,"unix");
		$nowsize = -s "$dir{conduit}/$gfsfile";
		print "as of $now, $dir{conduit}/$gfsfile is still arriving $nowsize < $expected{$file}{this} or $expected{$file}{full}\n";
		sleep 60;
		$itry++;
		$iret = &check_conduit_status($gfsfile);
	    }
	    $nowsize = -s "$dir{conduit}/$gfsfile";
	    print "after waiting, iret = $iret, itry = $itry, filesize = $nowsize ($expected{$file}{this} or $expected{$file}{full})\n";
	}
	$fstat = $iret;
	$infile = "$dir{conduit}/$gfsfile";
	if (-e "$rundir/.$gfsfile.ok" && $force != 1) {
	    $missing = "";
	} else {
	    &scan_wgrib2("$infile",$ihr);
	}
	if ($missing eq "") {
	    $now = &time_increment(time,"i4",0,0,"unix");
	    print "now have hour $ihr fields from conduit, $now\n";
	    system "touch $rundir/.$gfsfile.ok";
	    if ($onlymissing) {
		next;
	    } else {
		if ($nonsage) {
		    $com = "rsync -a ldm\@$freshairip:$infile $localfile";
		} else {
		    $com = "rsync -a $infile $localfile";
		}
		print "$com\n";
		system "$com";
		if ($localfile eq $sagefile) {
		    next;
		} else {
		    ##
		    ## our local version is good, let's rsync it to sage
		    ## even if domain name service is down
		    $com = "rsync -av $localfile mm5rt\@$sageip:$sagefile";
		    print "$com\n";
		    $iret = system "$com";
		    print "iret = $iret\n";
		    next;
		}
	    }
	} else {
	    print "missing $missing after conduit check\n";
	    $fstat = 1;
	}
    }
    if (! -e "$dir{conduit}/$gfsfile" || $fstat != 0) {
	$found{$gfsfile}{size} = 0;
	$found{$gfsfile}{loc} = "";
	foreach $subdir (@checkothers) {
	    next if ($skip{$subdir});
	    ##
	    ## there is a known problem here:
	    ##  if you get one file from the sage archive, you need to get all of
	    ##  the files from the sage archive -- the reason is that there are
	    ##  fewer vertical levels in the sage archive, or, you could
	    ##  force the preprocess.wrf.pl to run mod_levs.exe after metgrid.exe
	    ##  to get a consistent number of vertical levels.
	    ##  However, this mixed version really only occurred in my
	    ##  testing and should not occur in real-time.
	    ## 
	    ## if the file is here, make sure it isn't bad
	    if (-e "$rundir/$gfsfile") {
		&scan_wgrib2("$rundir/$gfsfile",$ihr);
		if ($missing ne "") {
		    print "$rundir/$gfsfile is not sufficient, removing it\n";
		    unlink "$rundir/$gfsfile";
		}
	    }
	    if (! -e "$rundir/$gfsfile") {
		chdir "$rundir";
		unlink "$rundir/.$gfsfile.ok" if (-e "$rundir/.$gfsfile.ok");
		##
		## we want to try to get the file from a remote server, but we don't
		## want to wait forever
		## we'll start a job in the background and monitor its progress
		## and will kill its process id if it takes too long
		$com = "$ENV{HOME}/bin/quick_gfs${res}_retrieval.$subdir.csh $ymdh $ihr";
		$iret = &run_forked("$com","$rundir/$gfsfile","$localstatusfile",$timelimit,1);
		if ($iret =~ /100%/) {
		    print "seems to have gotten the full file\n";
		} elsif ($iret =~ /\[(\d+)\/(\d+)\]/) {
		    $bytesreceived = $1;
		    $bytesexpected = $2;
		    if ($bytesreceived == $bytesexpected) {
			print "seems to have gotten the full bytes $bytesexpected\n";
		    } else {
			print "expected $bytesexpected but got $bytesreceived\n";
		    }
		} else {
		    print "$com did not seem to get the full file\n";
		}
	    }
	    $found{$gfsfile}{size} = -s "$rundir/$gfsfile";
	    if ($found{$gfsfile}{size} >=  40002921) {
		$found{$gfsfile}{loc} = "$rundir/$gfsfile";
		$infile = $found{$gfsfile}{loc};
		chdir "$rundir";
		if (-e "$rundir/.$gfsfile.ok" && $force != 1) {
		    $missing = "";
		} else {
		    &scan_wgrib2("$infile",$ihr);
		}
		if ($missing eq "") {
		    $now = &time_increment(time,"i4",0,0,"unix");
		    print "now have $ihr fields from $subdir, $now\n";
		    system "touch $rundir/.$gfsfile.ok";
		    if ($localfile ne $sagefile) {
			##
			## our local version is good, let's rsync it to sage
			## even if domain name service is down
			$com = "rsync -av $localfile mm5rt\@$sageip:$sagefile";
			print "$com\n";
			$iret = system "$com";
			print "iret = $iret\n";
		    }
		    if ($ENV{LOGNAME} =~ /mm5rt/ || $ENV{LOGNAME} eq "ldm") {
			$com = "rsync -av $localfile ldm\@$sageip:$dir{sageconduit}";
			print "$com\n";
			$iret = system "$com";
			print "iret = $iret\n";
		    }
		    last;
		} else {
		    print "still missing $missing after $subdir\n";
		}
	    }
	}
	if ($missing ne "") {
	    $now = &time_increment(time,"i4",0,0,"unix");
	    print "$now stilli missing hours $missing ... sleeping $sleep seconds\n";
	    system "sleep $sleep";
	    $retries++;
	    if ($retries > 4) {
		$retries = 0;
		$now = &time_increment(time,"i4",0,0,"unix");
		$message = "$0 tried 4 times to get $ihr for $ymdh, giving up at $now\n";
		&send_message($message,1,"","ERROR getting gfs$res for $ymdh $ihr from $0");
		system "rm /tmp/g*$$ > /dev/null 2>&1";
		exit 1;
	    }
	    goto RETRY;
	}
    } # end of checking files
#    print "why did i get here?\n";
}# end of the hours loop
$com = "ls -l $rundir/$ymdh/*$res*";
print "$com\n";
system "$com";
system "rm /tmp/g*$$";
$newdate = &time_increment(time,"i4",-11,0,"YYYYMMDDH12");
if ($thismachine ne "sage") {
    $end = &time_increment(time,"i4",-2,0,"YYYYMMDDH12");
} else {
    $end = &time_increment(time,"i4",-4,0,"YYYYMMDDH12");
}
if ($end > $ymdh) {
    $end = &time_increment("${ymdh}0000","yy",-2,0,"YYYYMMDDH12");
}
while ($newdate <= $end) {
    print "cleaning any $newdate files in $mainrundir/$newdate\n";
    if (-d "$mainrundir/$newdate") {
	$com = "/bin/rm -Rf $mainrundir/$newdate";
	print "$com\n";
	system "$com";
    }
    $newdate = &time_increment("${newdate}0000","yy",0,21600,"YYYYMMDDHH");
}
exit 0;

sub check_conduit_status {
    my ($file) = @_;
    if (-e "$dir{conduitstatus}/$file") {
	open(IN,"<$dir{conduitstatus}/$file");
	while (<IN>) {
	    if (/complete \((\d+) bytes/) {
		$expected{$file}{full} = $1;
	    } elsif (/nserted (\d+) of (\d+)/) {
		$expected{$file}{this} = $1;
		if ($expected{$file}{full} != $2) {
		    print "strange first status line says full size of $file is $expected{$file}{full}, second line says $2\n";
		}
		last;
	    }
	}
	close IN;
	if (-s "$dir{conduit}/$file" >= -s $expected{$file}{this}) {
	    return 0;
	} else {
	    return 1; # file is still arriving
	}
    } else {
	print "status file $dir{conduitstatus}/$file does not exist\n";
	return 2; # will not know full size, no status file exists
    }
} # end of subroutine check_conduit_status

sub scan_wgrib2 {
    my ($file,$ihr) = @_;
    my $fhr;
    if ($ihr == 0) {
	$fhr = "f00";
	@required_fields = @f00keys;
    } else {
	$fhr = "fhh";
	@required_fields = @fhhkeys;
    }
    my @fields;
    undef %found;
    $lastmtime = $mtime;
    open (FIELDS,"~/bin/wgrib2 $file | cut -f4-5 -d: |") || die "Cannot run wgrib2 $file";
    while (<FIELDS>) {
	chomp;
	if (/:/) {
	    $found{$fhr}{"$_"} = 1;
	}
    }
    close FIELDS;
    $missing = "";
    foreach $reqfld (@required_fields) {
	if (! $found{$fhr}{"$reqfld"}) {
	    $missing .= " $reqfld $ihr";
	}
    }
} # end of subroutine scan_wgrib2
    
sub load_wgrib2_protos {
    my $file,$fhr;
    foreach $split ("$mainrundir/wgrib2/gfs0p25.f00.proto,f00", "$mainrundir/wgrib2/gfs0p25.fhh.proto,fhh") {
	($file,$fhr) = split ',',$split;
	unless (open(IN,"<$file")) {
	    print "ERROR: cannot open $file for reading\n";
	    exit 1;
	}
	while (<IN>) {
	    next if (/^#/);
	    chomp;
	    if (/:/) {
		$protos{$fhr}{"$_"} = 1;
	    }
	}
	close IN;
    }
    @f00keys = keys(%{$protos{f00}});
    print "here is what we need at hour 00:\n@f00keys\n" if ($debug);
    @fhhkeys = keys(%{$protos{fhh}});
    print "here is what we need at all other hours:\n@fhhkeys\n" if ($debug);
} # end of subroutine load_wgrib2_protos

sub run_forked {
    my ($command,$localfile,$statusfile,$timelimit,$debug) = @_;
    my ($forked_pid,$negforked_pid,$commandstatusfile,$moresleep,$date,$ppid,$apid,$sumsleep,$start,$machine,$origcommand,$email_error_file,$email_error_file_timelimit);
    my @stats;
    my @in;
    $moresleep = -7; # will add 10 and start with 3 second sleep
    $timelimit = 480 if ($timelimit eq "");
    $email_error_file_timelimit = 3600;
    $mail_it   = 1 if ($mail_it eq "");
    $debug     = 0 if ($debug eq "");
    $sumsleep  = 0;
    $start     = time;
    $email_error_file = "/tmp/$ENV{LOGNAME}.$$.failure";
    $origcommand = $command;
    $commandstatusfile = "/tmp/run_and_monitor_a_command.$ENV{LOGNAME}.$$";
    $command .= " 1> $commandstatusfile 2>&1; echo \$? >> $commandstatusfile";
    
    unless ($forked_pid = fork) {
	##
	## fork returns zero, so I'm the child and I exec
	setpgrp(0,0);
	print "$command \n" if ($debug);
	exec "$command";
	exit; # terminate child processing here
    } else {
	##
	## parent now knows the child process id
	setpgrp($forked_pid,$forked_pid);
	$ppid = getpgrp $forked_pid;
	if ($debug) {
	    $date = &time_increment(time,"i4",0,0,"unix");
	    print "pid $$, while waiting for $command forked_pid = $forked_pid, pgrp $ppid to finish, I will monitor every $moresleep seconds to send email at $date\n";
	}
	until ( $ppid == -1 ) { ## wait until child is done
	    sleep 1;
	    $sumsleep++;
	    $apid = getpgrp $forked_pid;
	    if (&isrunning($forked_pid) == 1) {
		$ppid = $forked_pid;
	    } else {
		if (&isrunning($forked_pid) != 1) {
		    $ppid = -1;
		} else {
		    sleep 2;
		    $sumsleep += 2;
		}
	    }
	    if ($ppid != -1) {
		##
		## process $command is still running, but maybe it's out of
		## control, or the ps command is not working (this does happen)
		$moresleep += 10 if ($moresleep < $timelimit && $moresleep < 23);
		if ($debug) {
		    $date = &time_increment(time,"i4",0,0,"unix");
		    print "pid $$, sleeping $moresleep seconds waiting for $command to finish at $date, ppid=$ppid, forked_pid=$forked_pid, apid=$apid\n";
		}
		sleep $moresleep;
		$sumsleep += $moresleep;
		my $mytime = time;
		$date = &time_increment($mytime,"i4",0,0,"unix");
		my $size = -s "$localfile";
		open(OUT,">$statusfile");
		print OUT "#processid,user,thismachine,currentfilesize,i4time,unixtime\n";
		print OUT "$forked_pid,$ENV{LOGNAME},$thismachine,$size,$mytime,$date\n";
		close OUT;
		if ($debug) {
		    print "pid $$, after sleeping $moresleep seconds, it is now $date, incoming file $localfile is $size bytes\n";
		}
		if (time >= ($start + $timelimit)) {
		    $message = "ERROR in $0 @ARGV:\ntook too long (> $timelimit seconds) for $command, $date\n";
		    print "$message\n";
		    if ($forked_pid) {
			setpgrp($forked_pid,$forked_pid);
			$negforked_pid = -1 * $forked_pid;
			kill 9, $negforked_pid if ($negforked_pid != 0);
		    }
		    wait(); # reap children so don't have defunct, zombies
		    if (-e "$email_error_file") {
			@stats = stat($email_error_file);
			if ( (time - $stats[9]) > $email_error_file_timelimit) {
			    unlink "$email_error_file";
			    $mail_it = 1;
			    open(OUT,">$email_error_file");
			    print OUT "$message\n";
			    close OUT;
			} else {
			    $mail_it = 0;
			}
		    } else {
			$mail_it = 1;
			open(OUT,">$email_error_file");
			print OUT "$message\n";
			close OUT;
		    }
		    return "2,problem took too long to do the command $origcommand";
		}
	    } # end of checking ppid not = -1
	} # end of until ppid == -1
    } # end of waiting for fork

    if ($forked_pid) {
	setpgrp($forked_pid,$forked_pid);
	$negforked_pid = -1 * $forked_pid;
	kill 9, $negforked_pid if ($negforked_pid != 0);
    }
    wait(); # reap children so don't have defunct, zombies
    
    unless (open(IN,"<$commandstatusfile")) {
	print "$command never created $commandstatusfile\n";
	return "11,$command never created $commandstatusfile";
    } else {
	@in = <IN>;
	close IN;
	chomp($iret = $in[$#in]);
	unlink "$commandstatusfile";
    }
    print "full return from $origcommand remote_host = $remote_host was @in, iret = $iret\n" if ($debug);
    my $errormessage = "$iret";
    if ($#in > 0) {
	my $j;
	for ($j = 0; $j < $#in; $j++) {
	    chomp($errormessage .= ",$in[$j]");
	}
    }
    ##
    ## NOTE: by returning $errormessage, we get iret PLUS any error message
    ## in a form that still allows a simple 0 non-zero test on $iret
    unlink "$statusfile" if (-e "$statusfile");
    return "$errormessage"; 
} # end of subroutine run_forked


########################################################################
## from my date-lib.pl library
# Perl Time Libraries version 1.5  November 3, 2004
# Compiled or written by David Ovens, Univ. of Washington 
#                        (ovens@atmos.washington.edu)
# Permission to use and distribute these is granted in compliance with
# the GNU licensing agreement.  Let me know of any changes.
#------------------Perl Subroutine time_increment -------------------------
sub time_increment {
    my ($timein,$formatin,$incday,$incsec,$formatout,$timezone,$option) = @_;
    my ($timeout);
    local ($utc_local) = 8; # UTC is 8 hours later than Local Standard Time
#**********************************************************************c
#     time_increment will take various time inputs in UTC/GMT or local time
#         i4time
#         yyyymmddhhmmss
#         Julianday nseconds (space-separated)
#       add the requested number of days and/or seconds and return the
#       time in any of those forms or additional user-specified forms.    
#
#     INPUTS:
#       timein          variable      i4time [i4],
#                                     yyyymmddhhmmss [yy],
#                                     "Julianday nsec" [js],
#                                     "YYYY.MM.DD.HH:NN:SS" [yd],
#                                     "YYYYJJJHHNNSS" [yj],
#                                     "[YYYYMMDD][HH][:NN:SS]" [ye],
#                                     "YYYY-MM-DD HH:NN:SS" [ql]
#       formatin        CHARACTER     "i4",
#                                     "yy",
#                                     "js",
#                                     "yd",
#                                     "wrf",
#                                     "yj",
#                                     "ye",
#                                     "ql"
#       incday          INTEGER       Number of days to increment.
#       incsec          INTEGER       Number of seconds to increment.
#       formatout       CHARACTER     "i4",
#                                     "yy",
#                                     "js",
#                                     "yd",
#                                     "wrf",
#                                     "yj",
#                                     "ql"
#                                     "unix" = "day mon dd hh:nn:ss tzone year"
#                                       may use &quick_unix_time
#         NOTE:  yy alone is equivalent to yyyymmddhhnnss
#         Substitution will occur if formatout does not exactly match one 
#         of the 3 standards mentioned above.  Substitutions are:
#             year or yyyy   --> complete year (however many digits are needed)
#             yy             --> last 2 digits of year
#             mm             --> month as 2 digits
#             dd             --> day of month as 2 digits
#             d1             --> day of month as 1 or 2 digits
#             hh             --> hour as 2 digits
#             h12            --> nearest 12-hour time (00 or 12)
#             h1             --> hour as 1 or 2 digits, whichever fits
#             hN             --> nearest N-hour time (00 N 2N 3N ..)
#             nn             --> minute as 2 digits
#             ss             --> second as 2 digits
#             jjj            --> 3-digit day of the year (001 to 365 or 366)
#             jj             --> Astonomical Julian day (day 0 = 1/1/-4713)
#             jsec           --> number of seconds into Julian day
#             weekday        --> long name of day of week
#             day            --> short name of day of week
#             month          --> long name of month
#             mon            --> short name of month
#             ampm           --> indicate am/pm and use hours <= 12
#             tzone          --> 3-character time zone (DEFAULT is UTC,
#                                 unless optional timezone input is used)
#     OPTIONAL INPUTS:
#       timezone       CHARACTER   "local-UTC[,offset,name]",
#                                  "UTC-local[,offset,name]",
#                                  "local[,offset,name]",
#                                  "UTC", "PST", "HST",
#                                  "EST", "MST", "CST", "AST","AKST"
#                                   (DEFAULT is UTC)
#                                  The "local-UTC" and "UTC-local" timezone
#                                    choices are preferred.  They correctly 
#                                    convert local/UTC INPUT times to 
#                                    UTC/local OUTPUT times.  NOTE: that a
#                                    UTC-local conversion followed by a 
#                                    local-UTC conversion will be one hour off
#                                    between 1 and 2 AM local time on the date
#                                    of the October changeover.
#                                  NOTE: the optional ,offset,name can be added to
#                                        indicate the number of hours different from
#                                        UTC and the name of the standard of
#                                        standard/daylight time zones.
#                                        Examples:  MST same as utc-local,,mst
#                                        and UTC-LOCAL,-7,MST/MDT
#       NOTE                       i4 times are ALWAYS in UTC, gmtime() and
#                                    localtime() give appropriate values based
#                                    on this i4 time.
#                                  When "local" is chosen, the input time is
#                                    assumed to be the same time zone as 
#                                    timezone, though i4times will be 
#                                    automatically converted from their native
#                                    UTC values.  In other words, only cases
#                                    with formatin "i4" will do any time zone
#                                    conversions, others just assume inputs
#                                    and outputs are in the same time zone.
#       CAUTION                    Using "local" will give you DAYLIGHT time
#                                    during the 1 am to 2 am period of the
#                                    October changeover, since the function is
#                                    multivalued then.  Safer to use UTC-local.
#                                  Also, unless you are doing "UTC-local" or
#                                    "local-UTC", a time increment across a 
#                                    time zone switch behaves as though no 
#                                    switch occurs for all formatins EXCEPT 
#                                    "i4" which would have a one-hour shift in
#                                    one direction.
#       option          CHARACTER    "split" will split any output format 
#                                     containing a ${date}[._]${suffix} into
#                                     ${date},${suffix}.  This is useful for
#                                     full-valued master_pattern_file entries.
#     OUTPUT:
#       timeout         variable      Output time in chosen format.
#                                     -1 indicates an error. 
# Modifications:
#                D. Ovens (1/20/1999) Added upper-case to output format
#                                     so master_pattern_file can be used.
#                D. Ovens (4/7/1999) Changed output of yyyy or year from a
#                                     a 4-digit year to the complete year.
#                D. Ovens (4/20/1999) Added "yd" formats.
#                D. Ovens (5/3/1999) Added h12 (nearest 12-hour time) 
#                                     to output formats.
#                D. Ovens (8/26/1999) Added translation for (now,i4) 
#                                      formats for ($timein,$formatin).
#                                     Added yj input/output and jjj output 
#                                      formats.
#                D. Ovens (1/13/2000) Fixed it so a formatout "i4" with a
#                       timezone of "local-utc" will work.  The "i4" output
#                       still is not recommended with "local" or "utc-local".
#                D. Ovens (3/13/2000) Added hN (nearest N-hour time, where N 
#                                      is an integer) to output formats.
#                D. Ovens (1/3/2001) Fixed jjj output (day 4 now = 004).
#                D. Ovens (2/4/2001) Added form "ye" for expanded times.
#                D. Ovens (10/4/2001) Added form "ql" for mySQL datetime.
#                D. Ovens (10/10/2001) Changed "ye" so that a single digit
#                       expands to two:  1 means 01 not 1, 11, or 21.
#                D. Ovens (1/10/2002) Added form "unix" for formatout.
#                D. Ovens (4/4/2002) Added forms "h1", "d1", and "ampm".
#                D. Ovens (4/30/2015) Added form "wrf" for WRF datetime.
#**********************************************************************c
    my($yyyymmddhhmmss, $julianday, $nseconds, $ier,$newyear,
	  @months,@days,$year,$yy,$mm,$dd,$hh,$nn,$ss,$length,
	  $h12,$hx,$dot_location,$underscore_location,$time,$suffix,
	  $jday1,$juliandayofyear,$jjj,$incr,$count,$thisday,$inday,
          $h1,$d1);

#
#   Convert everything to Julianday nseconds format for calculations.
#
    if ($timezone ne "" && $timezone !~ /^(UTC|utc)$/) {
	$timezone =~ tr/a-z/A-Z/;
	my $dum;
	my $offset;
	my $tzonename;
	($dum,$offset,$tzonename) = split(',',$timezone);
	if ($timezone =~ /(EDT|EST)/) {
	    $utc_local = 5;
	    $dum = "UTC-LOCAL" if ($tzonename eq "");
	    $tzonename = "EST/EDT";
	} elsif ($timezone =~ /(CDT|CST)/) {
	    $utc_local = 6;
	    $dum = "UTC-LOCAL" if ($tzonename eq "");
	    $tzonename = "CST/CDT";
	} elsif ($timezone =~ /(PDT|PST)/) {
	    $dum = "UTC-LOCAL" if ($tzonename eq "");
	    $tzonename = "PST/PDT";
	} elsif ($timezone =~ /(MDT|MST)/) {
	    $utc_local = 7;
	    $dum = "UTC-LOCAL" if ($tzonename eq "");
	    $tzonename = "MST/MDT";
	} elsif ($timezone =~ /(NDT|NST)/) {
	    $utc_local = 3.5;
	    $dum = "UTC-LOCAL" if ($tzonename eq "");
	    $tzonename = "NST/NDT";
	} elsif ($timezone =~ /(AKDT|AKST)/) {
	    $utc_local = 9;
	    $dum = "UTC-LOCAL" if ($tzonename eq "");
	    $tzonename = "AKST/AKDT";
	} elsif ($timezone =~ /(ADT|AST)/) {
	    $utc_local = 4;
	    $dum = "UTC-LOCAL" if ($tzonename eq "");
	    $tzonename = "AST/ADT";
	} elsif ($timezone =~ /(HST)/) {
	    $utc_local = 10;
	    $dum = "UTC-LOCAL" if ($tzonename eq "");
	    $tzonename = "HST/HDT";
	} elsif ($timezone =~ /(WST)/) {
	    $utc_local = -8;
	    $dum = "UTC-LOCAL" if ($tzonename eq "");
	    $tzonename = "WST/WDT";
	} elsif ($offset ne "") {
	    $utc_local = -1 * $offset;
	    if ($tzonename eq "") {
		my $offsetp1 = $offset + 1;
		$tzonename = "UTC+$offset/UTC+$offsetp1";
	    }
	}
	$offset = -1 * $utc_local;
	$timezone = "$dum,$offset,$tzonename";
    }
    if ($formatout eq "unix" || $formatout eq "UNIX") {
	$formatout = "day mon dd hh:nn:ss tzone year";
	$timezone = "UTC-LOCAL" if ($timezone eq "");
	if ($formatin eq "i4") {
	    if ($timein =~ /^(NOW|now)/) {
		$timein = time;
	    }
	    $incsec += $incday*86400;
	    if ( (2147483647 - $timein) > $incsec && $timezone =~ /(utc-local|UTC-LOCAL)/) {
		return &quick_unix_time(($timein+$incsec));
	    }
	}
    }
    unless ($timezone) {
	$timezone = "UTC";
    }
    if ($formatin =~ /(i4|I4)/) {
	if ($timein =~ /^(NOW|now)/) {
	    $timein = time;
	}
	if ($timezone =~ /^(local|LOCAL)-(utc|UTC)/) {
	    $timezone = "UTC";
	}
	($julianday,$nseconds,$ier) = &jdaysec_i4time($timein);
    } elsif ($formatin =~ /(yy|YY)/) {
	$length = length($timein);
	$year = substr($timein,0,($length-10));
	$mm   = substr($timein,($length-10),2);
	$dd   = substr($timein,($length-8),2);
	$hh   = substr($timein,($length-6),2); 
	$nn   = substr($timein,($length-4),2);
	$ss   = substr($timein,($length-2),2);
	($julianday,$nseconds,$ier) = &jdaysec_yyyymmddhhmmss($year,$mm,$dd,$hh,$nn,$ss);
    } elsif ($formatin =~ /(yj|YJ)/) { # Y*JJJHHNNSS
	$length = length($timein);
	$year = substr($timein,0,($length-9));
	$jjj  = substr($timein,($length-9),3);
	$hh   = substr($timein,($length-6),2); 
	$nn   = substr($timein,($length-4),2);
	$ss   = substr($timein,($length-2),2);
	($julianday,$nseconds,$ier) = &jdaysec_yyyyjjjhhmmss($year,$jjj,$hh,$nn,$ss);
    } elsif ($formatin =~ /^(yd|YD|ql|QL|wrf|WRF)$/) {
	$length = length($timein);
	$year = substr($timein,0,($length-15));
	$mm   = substr($timein,($length-14),2);
	$dd   = substr($timein,($length-11),2);
	$hh   = substr($timein,($length-8),2); 
	$nn   = substr($timein,($length-5),2);
	$ss   = substr($timein,($length-2),2);
	($julianday,$nseconds,$ier) = &jdaysec_yyyymmddhhmmss($year,$mm,$dd,$hh,$nn,$ss);
    } elsif ($formatin =~ /(ye|YE)/) {
	($julianday,$nseconds,$ier) = &jdaysec_i4time(time);
	$yyyymmddhhnnss = &yyyymmddhhmmss_jdaysec($julianday,$nseconds);
	if ($timein ne "") {
	    $timein =~ s/^(\d+)?:?(\d\d)?:?(\d\d)?/$1$2$3/;
	    $ss = $3;
	    $nn = $2;
	    $dd = substr($1,-2);
	    if (length($1) == 1) {
		##
		## expand a single hour, n, to 0n
		$timein = "0$timein";
	    }
	    if ($ss ne "") {
		$incr = 1;
	    } elsif ($nn ne "") {
		$nseconds = $nseconds - ($nseconds%60);
		$incr = 60;
		$timein = "${timein}00";
	    } else {
		$nseconds = $nseconds - ($nseconds%3600);
		$incr = 3600;
		$timein = "${timein}0000";
	    }
	    $length = length($timein);
	    if ($length > 2 && $incr <= 1) {
		$nseconds = $nseconds - ($nseconds%60) + $ss;
		$incr = 60;
	    }
	    if ($length > 4 && $incr <= 60) {
		$nseconds = $nseconds - ($nseconds%3600) + $nn*60 + $ss;
		$incr = 3600;
	    }
	    if ($length > 6 && $incr <= 3600) {
		$nseconds = $nseconds - ($nseconds%86400) + $dd*3600 + $nn*60 + $ss;
		$inday   = substr($timein,0,(length($timein)-6));
		$thisday = substr($yyyymmddhhnnss,0,(length($yyyymmddhhnnss)-6));
		$thisday = substr($thisday,(-1*length($inday)));
		## is that complicated enough for you?
#	    print "thisday = $thisday  inday = $inday\n";
		if ($thisday ne $inday) {
		    $incr = 86400;
		}
	    }
	    if ($length < 14) {
		$count = 0;
		if ($length >= 12) {
		    ##
		    ## this allows future times but only in the current year
		    $year = substr($yyyymmddhhnnss,0,(length($yyyymmddhhnnss)-10));
		    $newyear = substr($yyyymmddhhnnss,0,(14-$length)) . substr($timein,0,($length-10));
		    until ($newyear <= $year &&
			   substr($newyear,(14-$length),($length-10)) eq substr($timein,0,($length-10)) ) {
			$newyear--;
		    }
		    $yyyymmddhhnnss = "$newyear" . substr($timein,($length-10),10);
		} else {
		    until ( substr($yyyymmddhhnnss,(-1*$length)) eq $timein) {
			$yyyymmddhhnnss = &time_increment("$julianday $nseconds","js",0,(-$incr*$count++),"yy");
#		print "$yyyymmddhhnnss $incr $count    $timein\n";
			if ($count > 200) {
			    print STDERR "ERROR in subroutine time_increment of date-lib.pl
Could not match $timein after 200 ${incr}-second time changes.\n";
			    return -1;
			}
		    }
		}
#	    print "got time in $count attempts\n";
	    } else {
#	    print "had sufficient info\n";
		$yyyymmddhhnnss = $timein;
	    }
	}
	$length = length($yyyymmddhhnnss);
	$year = substr($yyyymmddhhnnss,0,($length-10));
	$mm   = substr($yyyymmddhhnnss,($length-10),2);
	$dd   = substr($yyyymmddhhnnss,($length-8),2);
	$hh   = substr($yyyymmddhhnnss,($length-6),2); 
	$nn   = substr($yyyymmddhhnnss,($length-4),2);
	$ss   = substr($yyyymmddhhnnss,($length-2),2);
	($julianday,$nseconds,$ier) = &jdaysec_yyyymmddhhmmss($year,$mm,$dd,$hh,$nn,$ss);
    } elsif ($formatin =~ /(j(ulianday)?s|JS)/) {
	($julianday,$nseconds) = split(' ',$timein);
    }
#
#   Do the time increment.
#
    ($julianday,$nseconds) = &jdaysec_increment($julianday,$nseconds,$incday,$incsec);
#
#   Convert to output format.
#
    if ($formatout eq "i4" || $formatout eq "I4") {
	if ($timezone ne "UTC") {
	    ($julianday,$nseconds,$timezone) = &timezone_shift($julianday,$nseconds,$utc_local,$timezone);
	}
	$timeout = &i4time_jdaysec($julianday,$nseconds);
    } elsif ($formatout eq "js" || $formatout eq "JS") {
	if ($timezone ne "UTC") {
	    ($julianday,$nseconds,$timezone) = &timezone_shift($julianday,$nseconds,$utc_local,$timezone);
	}
	$timeout = "$julianday $nseconds";
    } elsif ($formatout =~ /^(yy|YY)(yymmddhhmmss|YYMMDDHHMMSS|yymmddhhnnss|YYMMDDHHNNSS)?$/) {
	if ($timezone ne "UTC") {
	    ($julianday,$nseconds,$timezone) = &timezone_shift($julianday,$nseconds,$utc_local,$timezone);
	}
	$timeout = &yyyymmddhhmmss_jdaysec($julianday,$nseconds);
    } elsif ($formatout =~ /^(yj|YJ)(yyyyjjjhhmmss|YYYYJJJHHMMSS|yyyyjjjhhnnss|YYYYJJJHHNNSS)?$/) {
	if ($timezone ne "UTC") {
	    ($julianday,$nseconds,$timezone) = &timezone_shift($julianday,$nseconds,$utc_local,$timezone);
	}
	$timeout = &yyyyjjjhhmmss_jdaysec($julianday,$nseconds);
    } elsif ($formatout =~ /^(yd|YD)$/) {
	if ($timezone ne "UTC") {
	    ($julianday,$nseconds,$timezone) = &timezone_shift($julianday,$nseconds,$utc_local,$timezone);
	}
	$yyyymmddhhmmss = &yyyymmddhhmmss_jdaysec($julianday,$nseconds);
	$length = length($yyyymmddhhmmss);
	$year = substr($yyyymmddhhmmss,0,($length-10));
	$mm   = substr($yyyymmddhhmmss,($length-10),2);
	$dd   = substr($yyyymmddhhmmss,($length-8),2);
	$hh   = substr($yyyymmddhhmmss,($length-6),2); 
	$nn   = substr($yyyymmddhhmmss,($length-4),2);
	$ss   = substr($yyyymmddhhmmss,($length-2),2);
	$timeout = sprintf("%s.%2.2d.%2.2d.%2.2d:%2.2d:%2.2d",
			   $year, $mm, $dd, $hh, $nn, $ss);
    } elsif ($formatout =~ /^(ql|QL)$/) {
	if ($timezone ne "UTC") {
	    ($julianday,$nseconds,$timezone) = &timezone_shift($julianday,$nseconds,$utc_local,$timezone);
	}
	$yyyymmddhhmmss = &yyyymmddhhmmss_jdaysec($julianday,$nseconds);
	$length = length($yyyymmddhhmmss);
	$year = substr($yyyymmddhhmmss,0,($length-10));
	$mm   = substr($yyyymmddhhmmss,($length-10),2);
	$dd   = substr($yyyymmddhhmmss,($length-8),2);
	$hh   = substr($yyyymmddhhmmss,($length-6),2); 
	$nn   = substr($yyyymmddhhmmss,($length-4),2);
	$ss   = substr($yyyymmddhhmmss,($length-2),2);
	$timeout = sprintf("%s-%2.2d-%2.2d %2.2d:%2.2d:%2.2d",
			   $year, $mm, $dd, $hh, $nn, $ss);
    } elsif ($formatout =~ /^(wrf|WRF)$/) {
	if ($timezone ne "UTC") {
	    ($julianday,$nseconds,$timezone) = &timezone_shift($julianday,$nseconds,$utc_local,$timezone);
	}
	$yyyymmddhhmmss = &yyyymmddhhmmss_jdaysec($julianday,$nseconds);
	$length = length($yyyymmddhhmmss);
	$year = substr($yyyymmddhhmmss,0,($length-10));
	$mm   = substr($yyyymmddhhmmss,($length-10),2);
	$dd   = substr($yyyymmddhhmmss,($length-8),2);
	$hh   = substr($yyyymmddhhmmss,($length-6),2); 
	$nn   = substr($yyyymmddhhmmss,($length-4),2);
	$ss   = substr($yyyymmddhhmmss,($length-2),2);
	$timeout = sprintf("%s-%2.2d-%2.2d_%2.2d:%2.2d:%2.2d",
			   $year, $mm, $dd, $hh, $nn, $ss);
    } elsif ($formatout eq "YYYYMMDDHH" || $formatout eq "yyyymmddhh") {
	if ($timezone ne "UTC") {
	    ($julianday,$nseconds,$timezone) = &timezone_shift($julianday,$nseconds,$utc_local,$timezone);
	}
	$timeout = &yyyymmddhhmmss_jdaysec($julianday,$nseconds);
	$timeout = substr($timeout,0,(length($timeout)-4));
    } else {
	if ($timezone ne "UTC") {
	    ($julianday,$nseconds,$timezone) = &timezone_shift($julianday,$nseconds,$utc_local,$timezone);
	}
	@months = ("", "January", "February", "March", "April", "May", "June",
		   "July", "August", "September", "October", "November", "December");
	@days = ("Sunday", "Monday", "Tuesday", "Wednesday", "Thursday",
		 "Friday", "Saturday");
	$yyyymmddhhmmss = &yyyymmddhhmmss_jdaysec($julianday,$nseconds);
	$length = length($yyyymmddhhmmss);
	$year = substr($yyyymmddhhmmss,0,($length-10));
	$yy   = substr($year,(length($year)-2),2);
	$mm   = substr($yyyymmddhhmmss,($length-10),2);
	$dd   = substr($yyyymmddhhmmss,($length-8),2);
	$d1   = $dd*1;
	$hh   = substr($yyyymmddhhmmss,($length-6),2);
	$nn   = substr($yyyymmddhhmmss,($length-4),2);
	$ss   = substr($yyyymmddhhmmss,($length-2),2);
	$h12  = sprintf("%2.2d",($hh - $hh % 12));
	$timeout = $formatout;
	$timeout =~ s/(yyyy|YYYY|year|YEAR)/$year/g;
	$timeout =~ s/(yy|YY)/$yy/g;
	$timeout =~ s/(mm|MM)/$mm/g;
	$timeout =~ s/(dd|DD)/$dd/g;
	$timeout =~ s/(d1|D1)/$d1/g;
	if ($timeout =~ /(ampm|AMPM)/) {
	    if ($hh eq "00") {
		$hh = $h12 = 12;
		$timeout =~ s/ampm/am/g;
		$timeout =~ s/AMPM/AM/g;
	    } elsif ($hh eq "12") {
		$timeout =~ s/ampm/pm/g;
		$timeout =~ s/AMPM/PM/g;
	    } elsif ($hh < 12) {
		$timeout =~ s/ampm/am/g;
		$timeout =~ s/AMPM/AM/g;
	    } else {
		$hh -= 12;
		$h12 -= 12;
		$hh = "0$hh" if ($hh < 10);
		$h12 = "0$h12" if ($h12 < 10);
		$timeout =~ s/ampm/pm/g;
		$timeout =~ s/AMPM/PM/g;
	    }
	}
	$timeout =~ s/(hh|HH)/$hh/g;
	$timeout =~ s/(h12|H12)/$h12/g;
	$h1      = $hh*1;
	$timeout =~ s/(h1|H1)/$h1/g;
	if ($timeout =~ /(h|H)(\d{1,2})/) {
	    $hx = sprintf("%2.2d",($hh - $hh % $2));
	    $timeout =~ s/(h\d{1,2}|H\d{1,2})/$hx/g;
	}
	$timeout =~ s/(nn|NN)/$nn/g;
	$timeout =~ s/(ss|SS)/$ss/g;
	if ($timeout =~ /(jjj|JJJ)/) {
	    $jday1 = &julday_jjjyyyy(1,$year);
	    $juliandayofyear = $julianday - $jday1 + 1;
	    $juliandayofyear = sprintf "%3.3d",$juliandayofyear;
	    $timeout =~ s/(jjj|JJJ)/$juliandayofyear/g;
	}
	$timeout =~ s/(jj|JJ)/$julianday/g;
	$timeout =~ s/(tzone|TZONE)/$timezone/g;
	$timeout =~ s/(jsec|JSEC)/$nseconds/g;
	$timeout =~ s/(month|MONTH|Month)/$months[$mm]/g;
	$timeout =~ s/(mon|MON|Mon)/substr($months[$mm],0,3)/e;
	if ($timeout =~ /(weekday|WEEKDAY|Weekday)/) {
	    $timeout =~ s/(weekday|WEEKDAY|Weekday)/$days[(($julianday+1)%7)]/;
	} elsif ($timeout =~ /(day|DAY|Day)/) {
	    $timeout =~ s/(day|DAY|Day)/substr($days[(($julianday+1)%7)],0,3)/e;
	}
##      now split time and suffix if requested
	if ($option eq "split") {
	    $dot_location = index($timeout,'.');
	    $underscore_location = index($timeout,'_');
	    if ($underscore_location > $dot_location) {
		if ($dot_location > 0) {
		    $time   = substr($timeout,0,$dot_location);
		    $suffix = substr($timeout,$dot_location); # from there to the end
		} else {
		    $time   = substr($timeout,0,$underscore_location);
		    $suffix = substr($timeout,$underscore_location); # from there to the end
		}
	    } elsif ($dot_location > $underscore_location) {
		if ($underscore_location > 0) {
		    $time   = substr($timeout,0,$underscore_location);
		    $suffix = substr($timeout,$underscore_location); # from there to the end
		} else {
		    $time   = substr($timeout,0,$dot_location);
		    $suffix = substr($timeout,$dot_location); # from there to the end
		}
	    } elsif ($dot_location == $underscore_location) {
		if ($dot_location > 0) {
		    print STDERR "ERROR in $0: date match does not make sense
$timeout has . and _ located at $dot_location and $underscore_location!\n";
		    exit 1;
		}
		$time = $timeout;
	    } # end of underscore and dot location checks
	    $timeout = join(',',$time,$suffix);
	} # end of option split
	$timeout =~ s/\\//g;
    }
    unless ($timeout) { $timeout = -1; }
    if ($ier) { $timeout = -1; }
    return $timeout;
} # end of subroutine time_increment

#------------------Perl Subroutine time_difference -------------------------
sub time_difference {
    my ($time1in,$format1in,$time2in,$format2in) = @_;
    my ($daydiff,$secdiff);
#**********************************************************************c
#     time_difference will take various time inputs in UTC/GMT for 2 
#       input times (time2 is assumed to be later than time1)
#         i4time
#         yyyymmddhhmmss
#         Julianday nseconds (space-separated)
#       and calculate the number of days and seconds between them
#
#     INPUTS:
#       time1in         variable      i4time,yyyymmddhhmmss,"Julianday nsec",
#                                       "YYYY.MM.DD.HH:NN:SS"
#       format1in       CHARACTER     "i4",  "yy", "js", "yd", "ql"
#       time2in         variable      i4time,yyyymmddhhmmss,"Julianday nsec",
#                                       "YYYY.MM.DD.HH:NN:SS"
#       format2in       CHARACTER     "i4",  "yy", "js", "yd", "ql"
#     OUTPUTS:
#       daydiff         integer       Number of days difference.
#                                     -1 indicates an error. 
#       secdiff         integer       Number of additional seconds separating 
#                                      time 2 from time 1.
#                                      
#  Modifications:
#     D. Ovens (4/7/1999) Changed yy input format to accept any length year.
#     D. Ovens (4/20/1999) Added yd formats.
#     D. Ovens (10/4/2001) Added ql format for mySQL datetime.
#**********************************************************************c
    my($julianday1, $nseconds1, $julianday2, $nseconds2, $ier,
	  $year, $mm, $dd, $hh, $nn, $ss, $length);
#
#   Convert everything to Julianday nseconds format for calculations.
#
    if ($format1in =~ /(i4|I4)/) {
	($julianday1,$nseconds1,$ier) = &jdaysec_i4time($time1in);
    } elsif ($format1in =~ /(yy|YY)/) {
	$length = length($time1in);
	$year = substr($time1in,0,($length-10));
	$mm   = substr($time1in,($length-10),2);
	$dd   = substr($time1in,($length-8),2);
	$hh   = substr($time1in,($length-6),2); 
	$nn   = substr($time1in,($length-4),2);
	$ss   = substr($time1in,($length-2),2);
	($julianday1,$nseconds1,$ier) = &jdaysec_yyyymmddhhmmss($year,$mm,$dd,$hh,$nn,$ss);
    } elsif ($format1in =~ /(yd|YD|ql|QL)/) {
	$length = length($time1in);
	$year = substr($time1in,0,($length-15));
	$mm   = substr($time1in,($length-14),2);
	$dd   = substr($time1in,($length-11),2);
	$hh   = substr($time1in,($length-8),2); 
	$nn   = substr($time1in,($length-5),2);
	$ss   = substr($time1in,($length-2),2);
	($julianday1,$nseconds1,$ier) = &jdaysec_yyyymmddhhmmss($year,$mm,$dd,$hh,$nn,$ss);
    } elsif ($format1in =~ /(j(ulianday)?s|JS)/) {
	($julianday1,$nseconds1) = split(' ',$time1in);
    }
    if ($format2in =~ /(i4|I4)/) {
	($julianday2,$nseconds2,$ier) = &jdaysec_i4time($time2in);
    } elsif ($format2in =~ /(yy|YY)/) {
	$length = length($time2in);
	$year = substr($time2in,0,($length-10));
	$mm   = substr($time2in,($length-10),2);
	$dd   = substr($time2in,($length-8),2);
	$hh   = substr($time2in,($length-6),2); 
	$nn   = substr($time2in,($length-4),2);
	$ss   = substr($time2in,($length-2),2);
	($julianday2,$nseconds2,$ier) = &jdaysec_yyyymmddhhmmss($year,$mm,$dd,$hh,$nn,$ss);
    } elsif ($format2in =~ /(yd|YD|ql|QL)/) {
	$length = length($time2in);
	$year = substr($time2in,0,($length-15));
	$mm   = substr($time2in,($length-14),2);
	$dd   = substr($time2in,($length-11),2);
	$hh   = substr($time2in,($length-8),2); 
	$nn   = substr($time2in,($length-5),2);
	$ss   = substr($time2in,($length-2),2);
	($julianday2,$nseconds2,$ier) = &jdaysec_yyyymmddhhmmss($year,$mm,$dd,$hh,$nn,$ss);
    } elsif ($format2in =~ /(j(ulianday)?s|JS)/) {
	($julianday2,$nseconds2) = split(' ',$time2in);
    }
#
#   Calculate the time difference.
#
    $daydiff = $julianday2 - $julianday1;
    $secdiff = $nseconds2 - $nseconds1;
    if ($daydiff > 0) { # we don't want negative seconds
	while ($secdiff >= 86400) {
	    $secdiff = $secdiff - 86400;
	    $daydiff = $daydiff + 1;
	}
	while ($secdiff < 0) {
	    $secdiff = $secdiff + 86400;
	    $daydiff = $daydiff - 1;
	}
    } elsif ($daydiff < 0) { # we don't want positive seconds
	while ($secdiff <= -86400) {
	    $secdiff = $secdiff + 86400;
	    $daydiff = $daydiff - 1;
	}
	while ($secdiff > 0) {
	    $secdiff = $secdiff - 86400;
	    $daydiff = $daydiff + 1;
	}
    } # else same day and seconds can be positive or negative
	
    return ($daydiff,$secdiff);
} # end of subroutine time_difference

#------------------Perl Subroutine jdaysec_yyyymmddhhmmss -----------------
sub jdaysec_yyyymmddhhmmss {
    my($year, $month, $day, $hour, $minute, $second) = @_;
    my($julianday, $nseconds, $ier);
    my(@ndays);
#**********************************************************************c
#     jdaysec_yyyymmddhhmmss returns the UTC/GMT Julian Day and number 
#       of seconds since midnight given 6 integers containing year,
#       month, day, hour, minute, second.
#
#
#     INPUTS:
#       year            INTEGER       FULL YEAR ONLY
#       month           INTEGER       1 to 12
#       day             INTEGER       1 to 31
#       hour            INTEGER       0 to 23
#       minute          INTEGER       0 to 59
#       second          INTEGER       0 to 59
#     OUTPUTS:
#       julianday       INTEGER       Full julian day.
#       nseconds        INTEGER       Number of seconds since midnight
#       ier             INTEGER       0 if successful, 1 if failure
#**********************************************************************c
    @ndays = (0,31,29,31,30,31,30,31,31,30,31,30,31);
    $nseconds = 0;
    $ier    = 0;

    if ($month < 1 || $month > 12 ||
	$day < 1 || $day > $ndays[$month] ||
	$hour < 0 || $hour > 23 ||
	$minute < 0 || $minute > 59 ||
	$second < 0 || $second > 59) {
# return error code.
#
	$ier = 1;
	print STDERR "error in jdaysec_yyyymmddhhmmss for $year.$month.$day.$hour:$minute:$second\n";
	print STDERR "please use a valid date\n";
    }

    $julianday = &julday_mmddyyyy($month,$day,$year);
    $nseconds = ($hour * 3600) + ($minute * 60) + $second;

    return ($julianday, $nseconds, $ier);
} # end of jdaysec_yyyymmddhhmmss

#------------------Perl Subroutine jdaysec_yyyyjjjhhmmss -----------------
sub jdaysec_yyyyjjjhhmmss {
    my($year, $jjj, $hour, $minute, $second) = @_;
    my($julianday, $nseconds, $ier);
    my(@ndays);
#**********************************************************************c
#     jdaysec_yyyyjjjhhmmss returns the UTC/GMT Julian Day and number 
#       of seconds since midnight given 5 integers containing year,
#       day of year, hour, minute, second.
#
#
#     INPUTS:
#       year            INTEGER       FULL YEAR ONLY
#       jjj             INTEGER       1 to 366
#       hour            INTEGER       0 to 23
#       minute          INTEGER       0 to 59
#       second          INTEGER       0 to 59
#     OUTPUTS:
#       julianday       INTEGER       Full julian day.
#       nseconds        INTEGER       Number of seconds since midnight
#       ier             INTEGER       0 if successful, 1 if failure
#**********************************************************************c
    $nseconds = 0;
    $ier    = 0;

    if ($jjj < 1 || $jjj > 366 ||
	$hour < 0 || $hour > 23 ||
	$minute < 0 || $minute > 59 ||
	$second < 0 || $second > 59) {
# return error code.
#
	$ier = 1;
	print STDERR "error in jdaysec_yyyyjjjhhmmss for $year.$jjj.$hour:$minute:$second\n";
	print STDERR "please use a valid date\n";
    }

    $julianday = &julday_jjjyyyy($jjj,$year);
    $nseconds = ($hour * 3600) + ($minute * 60) + $second;

    return ($julianday, $nseconds, $ier);
} # end of jdaysec_yyyyjjjhhmmss

#------------------Perl Subroutine yyyymmddhhmmss_jdaysec -----------------
sub yyyymmddhhmmss_jdaysec {
    my ($julianday,$nseconds) = @_;
    my ($yyyymmddhhmmss);
    my ($month, $day, $year, $hour, $minute, $second);
#**********************************************************************c
#     yyyymmddhhmmss_jdaysec returns the time as YYYYMMDDHHMMSS (a 14
#       character string) given an appropriate Julidan Day and number
#       of seconds since midnight
#
#     INPUTS:
#       julianday       INTEGER       Full Julian Day.
#       nseconds        INTEGER       Number of secons since midnight.
#     OUTPUTS:
#       yyyymmddhhmmss  CHAR*         Full time as character string
#                                      Year length varies as appropriate.
#     Modifications:
#       D. Ovens (4/7/1999) Changed output of so year is full length
#                            no matter what the year is.
#**********************************************************************c
    if ($nseconds >= 86400 || $nseconds < 0) { ($julianday,$nseconds) = &jdaysec_increment($julianday,0,0,$nseconds); }
    ($month,$day,$year) = &caldat($julianday);
    $hour = int($nseconds / 3600);
    $minute = int(($nseconds - $hour*3600) / 60);
    $second = int($nseconds - $hour*3600 - $minute*60);
    $yyyymmddhhmmss = sprintf("%d%2.2d%2.2d%2.2d%2.2d%2.2d",
			      $year, $month, $day, $hour, $minute, $second);
    return $yyyymmddhhmmss;
} # end of function yyyymmddhhmmss_jdaysec

#------------------Perl Subroutine yyyyjjjhhmmss_jdaysec -----------------
sub yyyyjjjhhmmss_jdaysec {
    my ($julianday,$nseconds) = @_;
    my ($yyyyjjjhhmmss);
    my ($jjj, $julianday1, $month, $day, $year, $hour, $minute, $second);
#**********************************************************************c
#     yyyyjjjhhmmss_jdaysec returns the time as YYYYJJJHHMMSS (a 13+
#       character string) given an appropriate Julidan Day and number
#       of seconds since midnight
#
#     INPUTS:
#       julianday       INTEGER       Full Julian Day.
#       nseconds        INTEGER       Number of secons since midnight.
#     OUTPUTS:
#       yyyyjjjhhmmss   CHAR*         Full time as character string
#                                      Year length varies as appropriate.
#     Modifications:
#       D. Ovens (8/26/1999) Created from yyyymmddhhmmss_jdaysec.
#**********************************************************************c
    if ($nseconds >= 86400 || $nseconds < 0) { ($julianday,$nseconds) = &jdaysec_increment($julianday,0,0,$nseconds); }
    ($month,$day,$year) = &caldat($julianday);
    $julianday1 = &julday_jjjyyyy(1,$year);
    $jjj  = $julianday - $julianday1 + 1;
    $hour = int($nseconds / 3600);
    $minute = int(($nseconds - $hour*3600) / 60);
    $second = int($nseconds - $hour*3600 - $minute*60);
    $yyyyjjjhhmmss = sprintf("%d%3.3d%2.2d%2.2d%2.2d",
			      $year, $jjj, $hour, $minute, $second);
    return $yyyyjjjhhmmss;
} # end of function yyyyjjjhhmmss_jdaysec

#------------------Perl Subroutine jdaysec_increment ----------------------
sub jdaysec_increment {
    my($jday,$nsec,$incday,$incsec) = @_;
    my($jdaynew,$nsecnew);
#**********************************************************************c
#     jdaysec_increment returns the UTC/GMT Julian Day and number 
#       of seconds since midnight given an initial Julian Day, number 
#       of seconds since midnight, number of days to increment, and
#       number of seconds to incremnt.
#
#
#     INPUTS:
#       jday            INTEGER       Full julian day.
#       nsec            INTEGER       Number of seconds since midnight
#       incday          INTEGER       Number of days to increment.
#       incsec          INTEGER       Number of seconds to increment.
#     OUTPUTS:
#       jdaynew         INTEGER       Full julian day.
#       nsecnew         INTEGER       Number of seconds since midnight
#**********************************************************************c
    $jdaynew = $jday + $incday;
    $nsecnew = $nsec + $incsec;
    while ($nsecnew >= 86400) {
	$nsecnew = $nsecnew - 86400;
	$jdaynew = $jdaynew + 1;
    }
    while ($nsecnew < 0) {
	$nsecnew = $nsecnew + 86400;
	$jdaynew = $jdaynew - 1;
    }
    return ($jdaynew,$nsecnew);
} #  end of subroutine jdaysec_increment

#------------------Perl Subroutine julday_mmddyyyy ------------------------
sub julday_mmddyyyy {
#**********************************************************************c
#     This is Numerical Recipes, JULDAY.  It returns the full Julian
#       day given three integers of the month, day, and year.
#     INPUTS:
#       mm              INTEGER       Month.
#       id              INTEGER       Day of the month.
#       iyyy            INTEGER       Full year.
#     OUTPUT: 
#       Julian Day      INTEGER       Full julian day
#**********************************************************************c
    my($mm,$id,$iyyy) = @_;
    my($igreg,$ja,$jm,$jy,$julday_mmddyyyy);

    $igreg = 15+31*(10+12*1582);
    if ($iyyy == 0) {
	print "There is no Year Zero.\n";
	exit(1);
    }
    if ($iyyy < 0) { $iyyy++; }
    if ($mm > 2) {
        $jy = $iyyy;
        $jm = $mm + 1;
    } else {
        $jy = $iyyy - 1;
        $jm = $mm + 13;
    }
    $julday_mmddyyyy = int(365.25*$jy) + int(30.6001*$jm) + $id + 1720995;
    if ( ($id+31*($mm+12*$iyyy)) >= $igreg ) {
        $ja = int(0.01*$jy);
        $julday_mmddyyyy = $julday_mmddyyyy+2-$ja+int(0.25*$ja);
    }
    return $julday_mmddyyyy;
} # end of function julday_mmddyyyy

#------------------Perl Subroutine julday_jjjyyyy ------------------------
sub julday_jjjyyyy {
#**********************************************************************c
#     This is patterned after Numerical Recipes, JULDAY.  It returns the 
#       full Julian day given the day of the year and the year.
#     INPUTS:
#       jjj             INTEGER       Day of the year 1 to 366.
#       iyyy            INTEGER       Full year.
#     OUTPUT: 
#       Julian Day      INTEGER       Full julian day
#**********************************************************************c
    my($jjj,$iyyy) = @_;
    my($igreg,$ja,$jm,$jy,$julday_jjjyyyy);

    $igreg = 15+31*(10+12*1582);
    if ($iyyy == 0) {
	print "There is no Year Zero.\n";
	exit(1);
    }
    if ($iyyy < 0) { $iyyy++; }
    $jy = $iyyy - 1;
    $jm = 14;
    $julday_jjjyyyy = int(365.25*$jy) + int(30.6001*$jm) + $jjj + 1720995;
    if ( ($jjj+31*(13*$iyyy)) >= $igreg ) {
        $ja = int(0.01*$jy);
        $julday_jjjyyyy = $julday_jjjyyyy+2-$ja+int(0.25*$ja);
    }
    return $julday_jjjyyyy;
} # end of function julday_jjjyyyy

#------------------Perl Subroutine caldat ------------------------------
sub caldat {
#**********************************************************************c
#     This is Numerical Recipes CALDAT, the complement of julday_mmddyyyy.
#       Given a full Julian day, this returns the month, day, and year.
#   INPUT:
#       julian              INTEGER       Month.
#   OUTPUTS:
#       month           INTEGER       Month.
#       day             INTEGER       Day of the month.
#       year            INTEGER       Full year.
#**********************************************************************c
    my($julian) = @_;
    my($igreg,$ja,$jalpha,$jb,$jc,$jd,$je,$month,$day,$year);

    $igreg = 2299161;
    if ($julian >= $igreg) {
	$jalpha = int((($julian-1867216)-0.25)/36524.25);
	$ja     = $julian+1+$jalpha-int(0.25*$jalpha);
    } else {
        $ja     = $julian;
    }
    $jb = $ja+1524;
    $jc = int(6680.+(($jb-2439870)-122.1)/365.25);
    $jd = 365*$jc+int(0.25*$jc);
    $je = int(($jb-$jd)/30.6001);
    $day = $jb-$jd-int(30.6001*$je);
    $month = $je - 1;
    if ($month > 12) { $month = $month - 12; }
    $year = $jc - 4715;
    if ($month > 2) { $year = $year - 1; }
    if ($year <= 0) { $year = $year - 1; }
    return ($month,$day,$year);
} # end of subroutine caldat

#------------------Perl Subroutine i4time_yyyymmddhhmmss------------------
sub i4time_yyyymmddhhmmss {
#     INPUTS 
    my($year, $month, $day, $hour, $minute, $second) = @_;
#     OUTPUTS
    my ($i4time, $ier);
#**********************************************************************c
#     i4time_yyyymmddhhmmss returns the UTC/GMT i4time (number of seconds
#       since midnight Jan 1, 1970) given 6 integers containing year,
#       month, day, hour, minute, second.
#
#     NOTE: this may break at 3:14:08 UTC Jan 19 2038 when integer*4
#       cannot handle i4time = 2147483648
#           
#
#     INPUTS:
#       year            INTEGER       FULL YEAR ONLY
#       month           INTEGER       1 to 12
#       day             INTEGER       1 to 31
#       hour            INTEGER       0 to 23
#       minute          INTEGER       0 to 59
#       second          INTEGER       0 to 59
#     OUTPUTS:
#       i4time          INTEGER       Number of seconds since 1/1/1970.
#       ier             INTEGER       0 if successful, 1 if failure
#**********************************************************************c
    my($juldaybase,@ndays);

    @ndays = (0,31,29,31,30,31,30,31,31,30,31,30,31);
    $juldaybase = 2440588; # Jan 1, 1970
    $ier    = 0;

    if ($month < 1 || $month > 12 ||
	$day < 1 || $day > $ndays[$month] ||
	$hour < 0 || $hour > 23 ||
	$minute < 0 || $minute > 59 ||
	$second < 0 || $second > 59) {
# return error code.
#
	$ier = 1;
	print "error in i4time_yyyymmddhhmmss\n";
    }

    $i4time = (&julday_mmddyyyy($month,$day,$year) - $juldaybase)*86400 +
	($hour * 3600) + ($minute * 60) + $second;
    return $i4time;
}  # End of perl subroutine i4time_yyyymmddhhmmss

#------------------Perl Subroutine yyyymmddhhmmss_i4time ---------------
sub yyyymmddhhmmss_i4time {
#**********************************************************************c
#     yyyymmddhhmmss_i4time returns fulltime in yyyymmddhhmmss
#       plus year, month, day, hour, minute, seconds, status, and isDST
#       given i4time (# of seconds since 00:00 01-jan-70)
#
#     NOTE: this will break at 3:14:08 UTC Jan 19 2038 when integer*4
#       cannot handle i4time = 2147483648
#
#     INPUTS:
#       i4time          INTEGER       Number of seconds since 01-jan-1970
#       timezone        CHARACTER     Time zone.
#     OUTPUTS: 
#       yyyymmddhhmmss  CHAR*         Full time as character string.
#       year            INTEGER       Full year (to the extent that i4time
#                                      can give it, see NOTE above).
#       month           INTEGER       1 to 12
#       day             INTEGER       1 to 31
#       hour            INTEGER       0 to 23
#       minute          INTEGER       0 to 59
#       second          INTEGER       0 to 59
#       ier             INTEGER       0 if successful, 1 if failure
#       isdst           INTEGER       1 if Daylight Savings Time, 0 otherwise
#
#     Modifications:
#           D. Ovens (4/7/1999) Changed output of year to format %d.
#
#           D. Ovens (1/13/2000) Fixed it to work beyond the 32-bit changeover,
#                                1/19/2038 at 3:14:07 UTC.  Now, only obsolete
#                                routines use gmtime() and localtime().
#           D. Ovens (1/22/2007) Changed the way &dstcalc is called.
#**********************************************************************c

    my($i4time,$timezone) = @_;
    my($julianday,$nseconds,$fulltime,$year,$month,$day,$hour,$min,$sec,$ier,$isdst) = "";
    local ($utc_local) = 8; # UTC is 8 hours later than Local Standard Time

    if (defined($i4time)) {
	$ier = 0;
    } else {
	$ier = 1;
	return ($fulltime,$year,$month,$day,$hour,$min,$sec,$ier,$isdst);
    }
    unless ($timezone) {
	$timezone = "UTC";
    }
    $timezone =~ tr/a-z/A-Z/;
    if ($timezone =~ /(EDT|EST)/) {
	$utc_local = 5;
    } elsif ($timezone =~ /(CDT|CST)/) {
	$utc_local = 6;
    } elsif ($timezone =~ /(MDT|MST)/) {
	$utc_local = 7;
    } elsif ($timezone =~ /(NDT|NST)/) {
	$utc_local = 3.5;
    } elsif ($timezone =~ /(AKDT|AKST)/) {
	$utc_local = 9;
    } elsif ($timezone =~ /(ADT|AST)/) {
	$utc_local = 4;
    } elsif ($timezone =~ /(HST)/) {
	$utc_local = 10;
    } elsif ($timezone =~ /(WST)/) {
	$utc_local = -8;
    }
    ($julianday,$nseconds,$ier) = &jdaysec_i4time($i4time);
    if ($timezone ne "UTC") {
	($julianday,$nseconds,$timezone) = &timezone_shift($julianday,$nseconds,$utc_local,$timezone);
    }
    $fulltime = &yyyymmddhhmmss_jdaysec($julianday,$nseconds);
    $length = length($fulltime);
    $year    = substr($fulltime,0,($length-10));
    $month   = substr($fulltime,($length-10),2);
    $day     = substr($fulltime,($length-8),2);
    $hour    = substr($fulltime,($length-6),2); 
    $min     = substr($fulltime,($length-4),2);
    $sec     = substr($fulltime,($length-2),2);
    $isdst   = &dstcalc($month,$day,$nseconds,$julianday);
    return ($fulltime,$year,$month,$day,$hour,$min,$sec,$ier,$isdst);
}       # End of perl subroutine yyyymmddhhmmss_i4time

#------------------Perl Subroutine yearmonthdayhour12------------------
sub yearmonthdayhour12 {
#**********************************************************************c
#     yearmonthdayhour12 is a quick way to convert an i4time (number of
#       seconds since 01-jan-1970) into a (yy)yymmddhh for the most
#       most recent 00Z/12Z time.
#
#     INPUTS:
#       i4time          INTEGER       Number of seconds since 01-jan-1970
#       timezone        CHARACTER     "gmt" or "local"
#       yeardigits      CHARACTER     "2" or anything else (not 2 means full)
#       hour12          INTEGER       0 (use the time as given by $i4time) 
#                                     1 (use the 0Z or 12Z time preceding $i4time)    
#     OUTPUT:
#      $yearmonthdayhour12     yymmddhh or yyyymmddhh based yeardigits
#                               with year being full value if yyyymmddhh chosen
#                              -1 (failure)
#     Modifications:
#           D. Ovens (4/7/1999) Changed output of year to format %d.
#**********************************************************************c

    my ($i4time,$timezone,$yeardigits,$hour12) = @_;
    my ($yyyymmddhhmmss,$year,$mon,$day,$hr,$length,
	   $julianday,$nsec);

##   Use jdaysec_increment to add i4time seconds to 
##   Jan 1, 1970 (Julianday 2440588)
    ($julianday,$nsec) = &jdaysec_increment(2440588,0,0,$i4time);
    $yyyymmddhhmmss = &yyyymmddhhmmss_jdaysec($julianday,$nsec);

    $length = length($yyyymmddhhmmss);
    $year = substr($yyyymmddhhmmss,0,($length-10));
    $mon  = substr($yyyymmddhhmmss,($length-10),2);
    $day  = substr($yyyymmddhhmmss,($length-8),2);
    $hr   = substr($yyyymmddhhmmss,($length-6),2); 
    if ($hour12) { $hr = $hr - $hr % 12; }
    if ($yeardigits eq "2") {
	$yearmonthdayhour12  = sprintf("%2.2d%2.2d%2.2d%2.2d",substr($year,(length($year)-2),2),$mon,$day,$hr);
    } else { # return full year by default
	$yearmonthdayhour12  = sprintf("%d%2.2d%2.2d%2.2d",$year,$mon,$day,$hr);
    }
    return $yearmonthdayhour12;
}       # End of perl subroutine yearmonthdayhour12

#------------------Perl Subroutine jdaysec_i4time -----------------------
sub jdaysec_i4time {
    my ($i4time) = @_;
    my($julianday, $nseconds, $ier);
#**********************************************************************c
#     jdaysec_i4time returns the UTC/GMT Julian Day and number 
#       of seconds since midnight given i4time (# of seconds since 
#       00:00 01-jan-70)
#
#     INPUT:
#       i4time          INTEGER       Number of seconds since 01-jan-1970
#     OUTPUTS:
#       julianday       INTEGER       Full julian day.
#       nseconds        INTEGER       Number of seconds since midnight
#       ier             INTEGER       0 if successful, 1 if failure
#**********************************************************************c

##   Use jdaysec_increment to add i4time seconds to 
##   Jan 1, 1970 (Julianday 2440588)
    $ier = 0;
    ($julianday,$nseconds) = &jdaysec_increment(2440588,0,0,$i4time);
    return ($julianday, $nseconds, $ier);
} # end of jdaysec_i4time

#------------------Perl Subroutine i4time_jdaysec -----------------------
sub i4time_jdaysec {
    my ($julianday, $nseconds) = @_;
    my($i4time, $ier);
#**********************************************************************c
#     i4time_jdaysec returns the i4time (UTC/GMT # of seconds since 
#       00:00 01-jan-70) given the Julian Day and number 
#       of seconds since midnight.
#     NOTE: this routine will output a number bigger than 2^31, so
#           be careful with any dates prior or 01-jan-1970 and
#           after 18-jan-2038 to make sure you can handle the big number
#
#     INPUT:
#       julianday       INTEGER       Full julian day.
#       nseconds        INTEGER       Number of seconds since midnight
#     OUTPUTS:
#       i4time          INTEGER*8     Number of seconds since 01-jan-1970
#       ier             INTEGER       0 if successful, 1 if failure
#**********************************************************************c

##   Calculate seconds since Jan 1, 1970 (Julianday 2440588)
##   32-bit changeover is Jan 18, 2038 at 3:14:07 UTC
##   (Julianday 2465443 11647 sec)
    $i4time = ($julianday - 2440588)*86400 + $nseconds;
    return ($i4time);
} # end of i4time_jdaysec

#------------------Perl Subroutine dstcalc ------------------------------
sub dstcalc {
#**********************************************************************c
#     This returns 1 if Daylight Savings Time is in effect and 0 if 
#       Standard time is in effect.
#     Based on first Sunday in April and last Sunday in October at 2 am
#       for years prior to 2007.  For 2007 and after the changes occur
#       on the second Sunday in March and the first Sunday in November.
#       Julianday 2454171 = March 11, 2007 first day of new dst.
#     INPUTS:
#       mm              INTEGER       Month.
#       dd              INTEGER       Day of the month.
#       ns              INTEGER       Seconds into the day.
#       dow             INTEGER       Day of week (0=Sunday, to 6=Saturday)
#     OPTIONAL INPUT:
#       timezone        INTEGER       hour difference UTC - local standard time
#                                      (DEFAULT is blank)
#                                     make calculations based on UTC time, 
#                                     which means shift at 10:00 UTC in April
#                                     and 9:00 UTC in October
#     OUTPUT: 
#                       INTEGER       1 = DST, 0 = not DST
#     Modifications:
#           D. Ovens (1/22/2007) Changed the way &dstcalc is called and
#             calculated due to the 2007 change in formulation -- fourth
#             argument was $dow = ($julianday+1)%7, now it is
#             $julianday so we can know the year.
#**********************************************************************c
    my($mm,$dd,$ns,$julianday,$timezone) = @_;
    my $dow = ($julianday+1)%7;

    if ($julianday < 2454171) {
	##
	## pre-2007 calculation
	if ($mm < 4 || $mm > 10) {
	    return 0;
	} elsif ($mm >= 5 && $mm <= 9) {
	    return 1;
	} elsif ($mm == 4) {
	    if ($dd > 7) {
		return 1;
	    } elsif ($dd >= ($dow+1)) {
		if ($timezone) {
		    if ($dow == 0 && $ns < (7200+$timezone*3600)) {
			return 0;
		    } else {
			return 1;
		    }
		} else {
		    if ($dow == 0 && $ns < 7200) {
			return 0;
		    } else {
			return 1;
		    }
		}
	    } else {
		return 0;
	    }
	} elsif ($mm == 10) {
	    if ($dd < 25) {
		return 1;
	    } elsif ($dd < ($dow+25)) {
		return 1;
	    } elsif ($dow == 0) {
		if ($timezone) { # UTC calculations
		    if ($ns < (7200+($timezone-1)*3600)) {
			return 1;
		    } else {
			return 0;
		    }
		} else { # local time calculations
		    if ($ns < 7200) {
			return 1;
		    } else {
			return 0;
		    }
		}
	    } else {
		return 0;
	    }
	} # end of month checks
    } else {
	##
	## 2007 and after calculations
	if ($mm < 3 || $mm > 11) {
	    return 0;
	} elsif ($mm >= 4 && $mm <= 10) {
	    return 1;
	} elsif ($mm == 3) {
	    if ($dd > 14) {
		## after the last possible second Sunday
		return 1;
	    } elsif ($dd >= ($dow+1+7)) {
		## in the second week or after (dd=8 to 14)
		if ($timezone) {
		    if ($dow == 0 && $ns < (7200+$timezone*3600)) {
			return 0;
		    } else {
			return 1;
		    }
		} else {
		    if ($dow == 0 && $ns < 7200) {
			return 0;
		    } else {
			return 1;
		    }
		}
	    } else {
		return 0;
	    }
	} elsif ($mm == 11) {
	    ## November fall back on first Sunday
	    if ($dd > 7) {
		## after the last possible first Sunday
		return 0;
	    } elsif ($dow == 0) {
		if ($timezone) { # UTC calculations
		    if ($ns < (7200+($timezone-1)*3600)) {
			return 1;
		    } else {
			return 0;
		    }
		} else { # local time calculations
		    if ($ns < 7200) {
			return 1;
		    } else {
			return 0;
		    }
		}
	    } elsif ($dd < ($dow + 1)) {
		## $dow = 0 on Sunday, if date is less than
		## the numerical day of the week (1-7),
		## then we cannot have reached the first Sunday
		## in November yet, so it's still DST.
		return 1;
	    } else {
		return 0;
	    }
	} # end of month checks
    }
} # end of subroutine dstcalc

#------------------Perl Subroutine timezone_shift -------------------------
sub timezone_shift {
    my ($julianday,$nseconds,$utc_local,$timezone) = @_;
#**********************************************************************c
#     timezone_shift will shift the given time to/from local/UTC
#       based on the value of timezone input
#
#     INPUTS:
#       julianday       INTEGER      Julian Day
#       nseconds        INTEGER      seconds into Julian Day
#       utc_local       INTEGER      number of hours for (UTC - local std time)
#       timezone        CHARACTER    "local-UTC","UTC-local","local","UTC"
#     OUTPUTS:
#       julianday       INTEGER      Julian Day
#       nseconds        INTEGER      seconds into Julian Day
#       timezone        CHARACTER    "UTC", "PST", "PDT"
#     Modifications:
#           D. Ovens (1/22/2007) Changed the way &dstcalc is called.
#**********************************************************************c
    my ($yyyymmddhhmmss,$length,$year,$yy,$mm,$dd,$hh,$nn,$ss);

    $yyyymmddhhmmss = &yyyymmddhhmmss_jdaysec($julianday,$nseconds);
    $length = length($yyyymmddhhmmss);
    $year = substr($yyyymmddhhmmss,0,($length-10));
    $yy   = substr($year,(length($year)-2),2);
    $mm   = substr($yyyymmddhhmmss,($length-10),2);
    $dd   = substr($yyyymmddhhmmss,($length-8),2);
    $hh   = substr($yyyymmddhhmmss,($length-6),2); 
    $nn   = substr($yyyymmddhhmmss,($length-4),2);
    $ss   = substr($yyyymmddhhmmss,($length-2),2);
    my $dum;
    my $offset;
    my $tzonename;
    my $standard = "PST";
    my $daylight = "PDT";
    ($dum,$offset,$tzonename) = split(',',$timezone);
    if ($tzonename ne "") {
	($standard,$daylight) = split('/',$tzonename);
    } elsif ($offset ne "") {
	$utc_local = -1 * $offset;
    }
    if ($timezone =~ /^(local|LOCAL)-(utc|UTC)/) {
	$timezone = "UTC";
	if ( &dstcalc($mm,$dd,$nseconds,$julianday) ) {
	    ($julianday,$nseconds) = &jdaysec_increment($julianday,$nseconds,0,($utc_local-1)*3600);
	} else {
	    ($julianday,$nseconds) = &jdaysec_increment($julianday,$nseconds,0,$utc_local*3600);
	}
    } elsif ($timezone =~ /^(UTC|utc)-(local|LOCAL)/) {
	if ( &dstcalc($mm,$dd,$nseconds,$julianday,$utc_local) ) {
	    $timezone = $daylight;
	    ($julianday,$nseconds) = &jdaysec_increment($julianday,$nseconds,0,($utc_local-1)*-3600);
	} else {
	    $timezone = $standard;
	    ($julianday,$nseconds) = &jdaysec_increment($julianday,$nseconds,0,$utc_local*-3600);
	}
    } elsif ($timezone =~ /^(loc(al)?|LOC(AL)?)/) {
	if ( &dstcalc($mm,$dd,$nseconds,$julianday) ) {
	    $timezone = $daylight;
	} else {
	    $timezone = $standard;
	}
    } # end of known timezone choices
    return ($julianday,$nseconds,$timezone);
} # end of subroutine timezone_shift

## This code is used to make time_increment(time,i4,0,0,"unix") run about
##  100 times faster!
## Since quick_unix_time relies on localtime, it may break
##     at 3:14:08 UTC Jan 19 2038 when integer*4 cannot handle 
##     i4time = 2147483648
## 
sub quick_unix_time {
    my @days = ("Sun", "Mon", "Tue", "Wed", "Thu",
		"Fri", "Sat");
    my @mons = ("Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec");
#
# NOTE: the following typo meant that I actually redefined $hour!!!
#    my $sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst,$year,$tzone;
# The parentheses are CRITICAL.    
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst,$tzone);
    ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = 
	localtime($_[0]);
    $year += 1900;
    if ($ENV{TZ} && $ENV{TZ} ne "US/Pacific") {
	$tzone = $ENV{TZ};
	$tzone = "$tzone Daylight" if ($isdst);
    } else {
	$tzone = "PST";
	$tzone = "PDT" if ($isdst);
    }
    return sprintf "%s %s %d %2.2d:%2.2d:%2.2d %s %d",$days[$wday],$mons[$mon],$mday,$hour,$min,$sec,$tzone,$year;
} # end of subroutine quick_unix_time

########## THE FOLLOWING SUBROUTINES ARE OBSOLETE, BUT MUST BE ###########
########## KEPT FOR COMPATIBILITY WITH OLD CODE.   D. Ovens 11/10/1998 ###
#
;#- Perl Subroutine changetime (cp 8/29/97 from ~marka/nexrad/chmin.pl) ------
sub changetime_min {

;# input       $intime  = Original yymmddhhmm
;# return      $outtime = Changed yymmddhhmm after adding $deltmin

      my ($intime, $deltmin) = @_;
      my ($yr,$mn,$dy,$hr,$min,@NDAYS,$minch,$hrch,$DYCH,$MNCH,$YRCH);

      $yr  = substr( $intime, 0, 2);
      $mn  = substr( $intime, 2, 2);
      $dy  = substr( $intime, 4, 2);
      $hr  = substr( $intime, 6, 2);
      $min = substr( $intime, 8, 2);

      @NDAYS = (0,31,28,31,30,31,30,31,31,30,31,30,31);

      if( $yr%4 == 0 && ($yr%100 != 0 || $yr%1000 == 0) ) { $NDAYS[2] = 29; }
      else { $NDAYS[2] = 28; }

      $minch = $min + $deltmin;
      $hrch  = $hr;
      $DYCH  = $dy;
      $MNCH  = $mn;
      $YRCH  = $yr;

      while( $minch < 0 || $minch > 59 ) {
         if( $minch < 0 ) {
            $minch += 60;
            $hrch--;
            if( $hrch == -1 ) {
               $hrch += 24;
               $DYCH--;
               if ($DYCH == 0) {
                  $MNCH--;
                  if ($MNCH == 0) {
                     $MNCH = 12;
                     $YRCH--;
                     if($YRCH%4 == 0 && ($YRCH%100 != 0 || $YRCH%1000 == 0)) {
                        $NDAYS[2] = 29;
                     }
                  }
                  $DYCH = $NDAYS[$MNCH];
               }
            }
         }
         else {      #  $minch > 59
            $minch -= 60;
            $hrch++;
            if( $hrch == 24 ) {
               $hrch -= 24;
               $DYCH++;
               if ($DYCH > $NDAYS[$MNCH]) {
                  $MNCH++;
                  if ($MNCH == 13) {
                     $MNCH  = 1;
                     $YRCH++;
                     if ($YRCH%4 == 0) { $NDAYS[2] = 29; }
                  }
                  $DYCH = 1;
               }
            }
         }
      }
      if( length($MNCH)  == 1 ) { $MNCH  = "0" . $MNCH; }
      if( length($DYCH)  == 1 ) { $DYCH  = "0" . $DYCH; }
      if( length($hrch)  == 1 ) { $hrch  = "0" . $hrch; }
      if( length($minch) == 1 ) { $minch = "0" . $minch; }

      return ( "$YRCH$MNCH$DYCH$hrch$minch" );
}
;#--------- End of Perl Subroutine changetime----------------------------------

;#- Perl Subroutine expandtime_min (cp 8/29/97 from ~marka/nexrad/expmin.pl)---
sub expandtime_min {
########################################################################c
#input     $timeArg        =>  Abbreviated form of the requested time
#                               in form yymmddhh:mm
#input     $offsetCurTime  =>  Offset in seconds from current time, used in
#                              calulating the default $time when $timeArg
#                              is specified as a null string.
#input     $yeardigits     =>  DEFAULT is last 2.
#input     $fulloutput     =>  Output time in yyyymmddhhnnss format. DEFAULT no
#output    $time           =>  Fully expanded form of the requested time in
#                               the form yymmddhhmm.
# Modifications:
#                D. Ovens (6/23/1999) Added fulloutput option.
########################################################################c
 my( $timeArg, $offsetCurTime, $yeardigits, $fulloutput ) = @_;
 my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst
	,$curmin,$curhour,$curday,$curmonth,$curyear
	,@dayspermonth,$selyear,$selmonth,$selday,$selhour,$selmin
	,$time,$i);

 if ($timeArg ne "") { $offsetCurTime = 0; }

 ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = 
                                               gmtime(time-$offsetCurTime);
 $curmin   = $min;
 $curhour  = $hour;
 $curday   = $mday;
 $curmonth = ++$mon;
 $curyear  = $year + 1900;

 @dayspermonth = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
 if ($curyear % 4 == 0 && ($curyear % 100 != 0 || $curyear % 1000 == 0)) { $dayspermonth[1] = 29; }

 if( $timeArg =~ /^(\d*)[,.;:]?(\d{0,2})$/ ) {
    $timeArg = $1;
    $selmin  = $2;
    if( length($selmin) == 0 ) {
       if( length($1) == 0 ) {
          $selmin = $curmin;
       }
       else { $selmin = "00"; }
    }
    if( length($selmin) == 1 ) { $selmin = "0" . $selmin; }
 }

 if (length($timeArg) == 1) { $timeArg = "0" . $timeArg; }
 if ($timeArg eq "00") { $timeArg = $mday . $timeArg; }
 for ($i=length($timeArg); $i<10; $i++) {   # zero fill to ten digits
    $timeArg = "0" . $timeArg;
 }

 $selyear  = substr( $timeArg, -10, 4 );
 $selmonth = substr( $timeArg, -6, 2 );
 $selday   = substr( $timeArg, -4, 2 );
 $selhour  = substr( $timeArg, -2, 2 );
 if ($selyear eq "0000") { $selyear = $curyear; }
 elsif ($selyear =~ /^00(\d\d)/) { $selyear = substr($curyear,0,2) . "$1"; }
 if ($selmonth == "00") {
    $selmonth = $curmonth;
    if ($selday eq "00") {
       $selday = $curday;
       if ($selhour eq "00") {
          $selhour = $curhour;
       }
       if ($selhour > $curhour) {
          $selday--;
          if ($selday == 0) {
             $selmonth--;
             if ($selmonth == 0) {
                $selmonth = 12;
                $selyear--;
             }
             $selday = $dayspermonth[$selmonth-1];
          }
       }
    }
    elsif ($selday > $curday) {
       $selmonth--;
       if ( $selmonth == 0 ) {
          $selmonth = 12;
          $selyear--;
       }
    }
 }
   
 if (length($selhour ) == 1) { $selhour  = "0" . $selhour; }
 if (length($selday  ) == 1) { $selday   = "0" . $selday; }
 if (length($selmonth) == 1) { $selmonth = "0" . $selmonth; }

 if ($yeardigits > 2) {
     if ($fulloutput) {
	 $time = "$selyear$selmonth$selday$selhour${selmin}00";
     } else {
	 $time = "$selyear$selmonth$selday$selhour$selmin";
     }
 } else {
     if ($fulloutput) {
	 $time = substr($selyear,2,2) . "$selmonth$selday$selhour${selmin}00";
     } else {
	 $time = substr($selyear,2,2) . "$selmonth$selday$selhour$selmin";
     }
 }
 return( "$time" );

}    # End Perl subroutine expandtime_min
;#-------------------------------------------------------------------------

#---------- Perl Subroutine calctime --------------------------------------
sub calctime {

# input      $answer   = Time series request for time
#            $year     = Default GMT year 
#            $mon      = Default GMT month
#            $day      = Default GMT day of month
#            $hour     = Default GMT hour
#            $year4d   = return 4 digit year or 2 digit year [optional]
# return     $yymmddhh = derived time value

      my( $answer, $year, $mon, $day, $hour, $year4d ) = @_;
      my($yy,$mm,$dd,$hh,$length);
      ($yy, $mm, $dd, $hh) = split( ' ', $answer );
      unless ($year4d) {
	  if ($yy) { $yy = $yy % 100; }
	  if ($year) { $year = $year % 100; }
      }
      if( ($length = length($hh)) < 2 ) {
         if( $length == 0 ) {
             $hh = $hour;
             $length = length($hour);
         } 
         if( $length == 1 ) { $hh = "0" . $hh; }
      }
      if( ($length = length($dd)) < 2 ) {
         if( $length == 0 ) {
             $dd = $day;
             $length = length($day);
         } 
         if( $length == 1 ) { $dd = "0" . $dd; }
      }
      if( ($length = length($mm)) < 2 ) {
         if( $length == 0 ) {
             $mm = $mon;
             $length = length($mon);
         } 
         if( $length == 1 ) { $mm = "0" . $mm; }
      }
      if( ($length = length($yy)) < 2 ) {
         if( $length == 0 ) {
             $yy = $year;
             $length = length($year);
         } 
         if( $length == 1 ) { $yy = "0" . $yy; }
      }

      if ($year4d) {
	  $yy = 1900 + $yy;
	  return sprintf ("%4.4d%2.2d%2.2d%2.2d",$yy,$mm,$dd,$hh);
      } else {
	  return ( "$yy$mm$dd$hh" );
      }
}
#--------- End of subroutine calctime----------------------------------
   
sub changetime10 {
# inputs:       $timein  = YYYYMMDDHH
#               $changehour = number of hours to change $timein
# output:       $timeout = YYYYMMDDHH  ($timein + $changehour)
    my($timein,$changehour) = @_;
    my($timeout);
    $timeout = &time_increment("${timein}0000","yy",0,($changehour*3600),"YYYYMMDDHH");
    return $timeout;
}
#--------- End of subroutine changetime10

sub expandtime10 {

#input     $timeArg        =>  Abbreviated form of the requested time
#input     $offsetCurTime  =>  Offset in seconds from current time, used in
#                              calulating the default $time when $timeArg
#                              is specified as a null string.
#return value              =>  Fully expanded form of the requested time in
#                              the form yyyymmddhh.

 my( $timeArg, $offsetCurTime ) = @_;
 my( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst );
 my( $selyear, $selmonth, $selday, $selhour );
 my( $curyear, $curmonth, $curday, $curhour, @dayspermonth, $yy );

 if ($timeArg ne "") { $offsetCurTime = 0; }

 ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = 
                                               gmtime(time-$offsetCurTime);
 $curhour  = $hour;
 $curday   = $mday;
 $curmonth = ++$mon;
 $curyear  = $year + 1900;

 @dayspermonth = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
 if ($curyear % 4 == 0) { $dayspermonth[1] = 29; }

 if( length($timeArg) == 1 ) { $timeArg = "0" . $timeArg; }
 if( length($timeArg) == 8 ) {
    $yy = substr($timeArg, 0, 2);
    if( $yy > 50 ) { $timeArg = "19$timeArg"; }
    else { $timeArg = "20$timeArg"; }
 } 
 if ($timeArg eq "00") { $timeArg = $mday . $timeArg; }
 for ( my $i=length($timeArg); $i<10; $i++) {   # zero fill to ten digits
    $timeArg = "0" . $timeArg;
 }
 $selyear  = substr( $timeArg, -10, 4 );
 $selmonth = substr( $timeArg, -6, 2 );
 $selday   = substr( $timeArg, -4, 2 );
 $selhour  = substr( $timeArg, -2, 2 );
 if ($selyear eq "0000") { $selyear = $curyear; }
 if ($selmonth == "00") {
    $selmonth = $curmonth;
    if ($selday eq "00") {
       $selday = $curday;
       if ($selhour eq "00") {
          $selhour = $curhour;
       }
       if ($selhour > $curhour) {
          $selday--;
          if ($selday == 0) {
             $selmonth--;
             if ($selmonth == 0) {
                $selmonth = 12;
                $selyear--;
             }
             $selday = $dayspermonth[$selmonth-1];
          }
       }
    }
    elsif ($selday > $curday) {
       $selmonth--;
       if ( $selmonth == 0 ) {
          $selmonth = 12;
          $selyear--;
       }
    }
 }
   
 if (length($selhour ) == 1) { $selhour  = "0" . $selhour; }
 if (length($selday  ) == 1) { $selday   = "0" . $selday; }
 if (length($selmonth) == 1) { $selmonth = "0" . $selmonth; }

 return "$selyear$selmonth$selday$selhour";

}
#--------------- End of subroutine expandtime10 --------------

sub changetime {

# input       $intime  = Original yymmddhh
# return      $outtime = Changed yymmddhh after adding $delthr 

      my ($intime, $delthr) = @_;
      my ($yr,$mn,$dy,$ho,@NDAYS,$HRCH,$DYCH,$MNCH,$YRCH);

      $yr = substr( $intime, 0, 2);
      $mn = substr( $intime, 2, 2);
      $dy = substr( $intime, 4, 2);
      $ho = substr( $intime, 6, 2);

      @NDAYS = (0,31,28,31,30,31,30,31,31,30,31,30,31);

      if ($yr%4 == 0) { $NDAYS[2] = 29; }
      else { $NDAYS[2] = 28; }

      $HRCH = $ho + $delthr;
      $DYCH = $dy;
      $MNCH = $mn;
      $YRCH = $yr;

      while ($HRCH < 0 || $HRCH > 23) {
         if ($HRCH < 0) {
            $HRCH += 24;
            $DYCH--;
            if ($DYCH == 0) {
               $MNCH--;
               if ($MNCH == 0) {
                  $MNCH = 12;
                  $YRCH--;
                  if ($YRCH%4 == 0) { $NDAYS[2] = 29; }
               }
               $DYCH = $NDAYS[$MNCH];
            }
         }
         if ($HRCH > 23) {
            $HRCH -= 24;
            $DYCH++;
            if ($DYCH > $NDAYS[$MNCH]) {
               $MNCH++;
               if ($MNCH == 13) {
                  $MNCH  = 1;
                  $YRCH++;
                  if ($YRCH%4 == 0) { $NDAYS[2] = 29; }
               }
               $DYCH = 1;
            }
         }
      }
      if( length($MNCH) == 1 ) { $MNCH = "0" . $MNCH; }
      if( length($DYCH) == 1 ) { $DYCH = "0" . $DYCH; }
      if( length($HRCH) == 1 ) { $HRCH = "0" . $HRCH; }

      return ( "$YRCH$MNCH$DYCH$HRCH" );
}
#--------- End of Perl Subroutine changetime----------------------------------

#------------------Perl Subroutine expandtime------------------------------
sub expandtime {

#input     $timeArg        =>  Abbreviated form of the requested time
#input     $offsetCurTime  =>  Offset in seconds from current time, used in
#                              calulating the default $time when $timeArg
#                              is specified as a null string.
#output    $time           =>  Fully expanded form of the requested time in
#                               the form yymmddhh.

 my( $timeArg, $offsetCurTime ) = @_;
 my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst
	,$curhour,$curday,$curmonth,$curyear
	,@dayspermonth,$selyear,$selmonth,$selday,$selhour
	,$i,$time);


 if ($timeArg ne "") { $offsetCurTime = 0; }

 ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = 
                                               gmtime(time-$offsetCurTime);
 $curhour  = $hour;
 $curday   = $mday;
 $curmonth = ++$mon;
 $curyear  = $year;

 @dayspermonth = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
 if ((1900+$curyear) % 4 == 0) { $dayspermonth[1] = 29; }

 if (length($timeArg) == 1) { $timeArg = "0" . $timeArg; }
 if ($timeArg eq "00") { $timeArg = $mday . $timeArg; }
 for ($i=length($timeArg); $i<8; $i++) {   # zero fill to eight digits
    $timeArg = "0" . $timeArg;
 }
 $selyear  = substr( $timeArg, -8, 2 );
 $selmonth = substr( $timeArg, -6, 2 );
 $selday   = substr( $timeArg, -4, 2 );
 $selhour  = substr( $timeArg, -2, 2 );
 if ($selyear eq "00") { $selyear = $curyear; }
 if ($selmonth == "00") {
    $selmonth = $curmonth;
    if ($selday eq "00") {
       $selday = $curday;
       if ($selhour eq "00") {
          $selhour = $curhour;
       }
       if ($selhour > $curhour) {
          $selday--;
          if ($selday == 0) {
             $selmonth--;
             if ($selmonth == 0) {
                $selmonth = 12;
                $selyear--;
             }
             $selday = $dayspermonth[$selmonth-1];
          }
       }
    }
    elsif ($selday > $curday) {
       $selmonth--;
       if ( $selmonth == 0 ) {
          $selmonth = 12;
          $selyear--;
       }
    }
 }
   
 if (length($selhour ) == 1) { $selhour  = "0" . $selhour; }
 if (length($selday  ) == 1) { $selday   = "0" . $selday; }
 if (length($selmonth) == 1) { $selmonth = "0" . $selmonth; }

 $time = "$selyear$selmonth$selday$selhour";
 return( "$time" );

}    # End Perl subroutine expandtime
#-------------------------------------------------------------------------
## end of date-lib.pl
########################################################################

########################################################################
## from my local-lib.pl
########################### printallvars ############################
## Added 8/20/2015 by David Ovens
## This prints all variables that have been defined in the current scope
##  except $_ and $usage and strange $_... variables.
## Pass it a non-empty argument to see everything.
## Usage: &printallvars();
## Very handy for debugging.
sub printallvars {
    my ($all) = @_;
    foreach my $entry ( sort(keys %main::) ) {
        $entry =~ s/([\x00-\x20])/sprintf("0x%x",unpack("C",$1))/eg;
        $entry =~ s/([\x7f-\xff])/sprintf("0x%x",unpack("C",$1))/eg;
	unless ($all)  {
	    next if ($$entry eq undef || $entry eq "usage" || $entry =~ /^_/);
	}
        printf "%26s ",$entry;
        print (($$entry eq undef) ? "undefined\n" :  ($$entry,"\n"));
    }
} # end of subroutine printallvars
########################### printallvars ############################

################## run_and_monitor_a_command ############################
## Added: September 4, 2012 by David Ovens
## This is the best way I've found to run and monitor the progress of
## a command.  I don't set alarms, but fork and monitor the process id.
## Arguments:
##   required:   $command     --> what to run
##   optional:   $timelimit   --> seconds to wait before giving up and notifying of error
##                                DEFAULT = 3600
##               $remote_host --> the name of a remote machine, if used
##                                DEFAULT = blank, tries to figure out from $command
##               $mail_it     --> 0 = don't mail error message
##                                DEFAULT = 1
##                                NOTE: To avoid email floods, this program
##                                will only send a specific email every
##                                $timelimit seconds.
##               $debug       --> print debug information
##                                DEFAULT = 0
##               $warningtime --> seconds to wait before sending a warning email
##                                DEFAULT = same as $timelimit
sub run_and_monitor_a_command {
    my ($command,$timelimit,$remote_host,$mail_it,$debug,$warningtime) = @_;
    my ($forked_pid,$negforked_pid,$statusfile,$moresleep,$date,$ppid,$apid,$sumsleep,$thismachine,$start,$machine,$origcommand,$email_error_file,$email_error_file_timelimit);
    my @stats;
    my @in;
    $moresleep = -7; # will add 10 and start with 3 second sleep
    $timelimit = 3600 if ($timelimit eq "");
    $email_error_file_timelimit = 3600;
    $mail_it   = 1 if ($mail_it eq "");
    $debug     = 0 if ($debug eq "");
    $warningtime = $timelimit if ($warningtime eq "");
    $sumsleep  = 0;
    $thismachine = &gethost();
    $start     = time;
    $statusfile = "/tmp/run_and_monitor_a_command.$ENV{LOGNAME}.$$";
    ##
    ## basic attempts to try to determine remote_host
    ## It can be pretty complicated, so I've added the 3rd argument
    if ($remote_host eq "") {
	$remote_host = $thismachine;
	if ($command =~ /(ssh|rsh|rsync|ftp|\@)/) {
	    if ($command =~ /\@(\S+)/) {
		$remote_host = $1;
	    } elsif ($command =~ /rsync.*\s(\S+):/) {
		$remote_host = $1;
	    } elsif ($command =~ /(\w+\.\w+\.\w+\.\w+)/) {
		$remote_host = $1;
	    } elsif ($command =~ /(ssh|rsh)\s(\S+)\s/) {
		$remote_host = $2;
	    }
	} elsif ($command =~ /wget.*\/\/(\S+)/) {
	    $remote_host = $1;
	}
    }
    $email_error_file = "/tmp/$ENV{LOGNAME}.$remote_host.failure";
    $origcommand = $command;
    $command .= " 1> $statusfile 2>&1; echo \$? >> $statusfile";
    
    unless ($forked_pid = fork) {
	##
	## fork returns zero, so I'm the child and I exec
	setpgrp(0,0);
	print "$command \n" if ($debug);
	exec "$command";
	exit; # terminate child processing here
    } else {
	##
	## parent now knows the child process id
	setpgrp($forked_pid,$forked_pid);
	$ppid = getpgrp $forked_pid;
	if ($debug) {
	    $date = &time_increment(time,"i4",0,0,"unix");
	    print "pid $$, while waiting for $command forked_pid = $forked_pid, pgrp $ppid to finish, I will monitor every $moresleep seconds to send email at $date\n";
	}
	until ( $ppid == -1 ) { ## wait until child is done
	    sleep 1;
	    $sumsleep++;
	    $apid = getpgrp $forked_pid;
	    if (&isrunning($forked_pid) == 1) {
		$ppid = $forked_pid;
	    } else {
		if (&isrunning($forked_pid) != 1) {
		    $ppid = -1;
		} else {
		    sleep 2;
		    $sumsleep += 2;
		}
	    }
	    if ($ppid != -1) {
		##
		## process $command is still running, but maybe it's out of
		## control, or the ps command is not working (this does happen)
		$moresleep += 10 if ($moresleep < $timelimit);
		if ($debug) {
		    $date = &time_increment(time,"i4",0,0,"unix");
		    print "pid $$, sleeping $moresleep seconds waiting for $command to finish at $date, ppid=$ppid, forked_pid=$forked_pid, apid=$apid\n";
		}
		sleep $moresleep;
		$sumsleep += $moresleep;
		$date = &time_increment(time,"i4",0,0,"unix");
		if ($debug) {
		    print "pid $$, after sleeping $moresleep seconds, it is now $date\n";
		}
		if (time >= ($start + $timelimit)) {
		    $message = "ERROR in $0 @ARGV:\ntook too long (> $timelimit seconds) for $command, REAL PROBLEM with $remote_host, ABORTING $0 @ARGV on $thismachine as of $date\n";
		    print "$message\n";

		    if ($forked_pid) {
			setpgrp($forked_pid,$forked_pid);
			$negforked_pid = -1 * $forked_pid;
			kill 9, $negforked_pid if ($negforked_pid != 0);
		    }
		    wait(); # reap children so don't have defunct, zombies
		    if (-e "$email_error_file") {
			@stats = stat($email_error_file);
			if ( (time - $stats[9]) > $email_error_file_timelimit) {
			    unlink "$email_error_file";
			    $mail_it = 1;
			    open(OUT,">$email_error_file");
			    print OUT "$message\n";
			    close OUT;
			} else {
			    $mail_it = 0;
			}
		    } else {
			$mail_it = 1;
			open(OUT,">$email_error_file");
			print OUT "$message\n";
			close OUT;
		    }
		    &send_message($message,$mail_it,"","PROBLEM with local or $remote_host from $0");
		    return "2,problem with local or $remote_host because it took too long to do the command $origcommand";
		}
		if ((time - $start) > $warningtime) {
		    $message = "WARNING in $0 @ARGV:\ntaking too long (> $warningtime seconds) for $command, POSSIBLE PROBLEM with a $remote_host\n";
		    print "$message\n";
		    if (-e "$email_error_file") {
			@stats = stat($email_error_file);
			if ( (time - $stats[9]) > $email_error_file_timelimit) {
			    unlink "$email_error_file";
			    $mail_it = 1;
			    open(OUT,">$email_error_file");
			    print OUT "$message\n";
			    close OUT;
			} else {
			    $mail_it = 0;
			}
		    } else {
			$mail_it = 1;
			open(OUT,">$email_error_file");
			print OUT "$message\n";
			close OUT;
		    }
		    &send_message($message,1,"","POSSIBLE PROBLEM with $remote_host from $0");
		    $start = time; # reset that warning alarm
		}
	    } # end of checking ppid not = -1
	} # end of until ppid == -1
    } # end of waiting for fork

    if ($forked_pid) {
	setpgrp($forked_pid,$forked_pid);
	$negforked_pid = -1 * $forked_pid;
	kill 9, $negforked_pid if ($negforked_pid != 0);
    }
    wait(); # reap children so don't have defunct, zombies
    
    unless (open(IN,"<$statusfile")) {
	print "$command never created $statusfile\n";
	return "11,$command never created $statusfile";
    } else {
	@in = <IN>;
	close IN;
	chomp($iret = $in[$#in]);
	unlink "$statusfile";
    }
    print "full return from $origcommand remote_host = $remote_host was @in, iret = $iret\n" if ($debug);
    my $errormessage = "$iret";
    if ($#in > 0) {
	my $j;
	for ($j = 0; $j < $#in; $j++) {
	    chomp($errormessage .= ",$in[$j]");
	}
    }
    ##
    ## NOTE: by returning $errormessage, we get iret PLUS any error message
    ## in a form that still allows a simple 0 non-zero test on $iret
    return "$errormessage"; 
} # end of subroutine run_and_monitor_a_command
################## run_and_monitor_a_command ############################

############################ gethost ###################################
## A faster way to get the short machine name
## OPTIONAL input: "longname", "fullname", "ip"
sub gethost {
    ##
    ## step one, get all IP addresses of host, store in @ips and @locals
    my($option,$othermachine) = @_;
    use Socket;
    use Sys::Hostname;
    my $addr;
    my $thismachine;
    my $iaddr;
    if ($othermachine eq "") {
	$thismachine = hostname();
	return $thismachine if ($option eq ""); # added 5/3/2021
	$addr = inet_ntoa(scalar(gethostbyname($thismachine)) || 'localhost');
    } else {
	$thismachine = $othermachine;
	$addr = inet_ntoa(scalar(gethostbyname($thismachine)));
    }
    return $addr if ($option eq "ip");
    if ($option =~ /^(long|full)name/) {
	if ($othermachine eq "") {
	    return $thismachine;
	} else {
	    $iaddr = inet_aton($addr);
	    return gethostbyaddr($iaddr, AF_INET);
	}
    } else {
	# short name of this machine
	if ($thismachine !~ /^\d/) {
	    $thismachine =~ s/(\w+)(\..*)?/$1/;
	} else {
	    $iaddr = inet_aton($addr);
	    $thismachine = gethostbyaddr($iaddr, AF_INET);
	    $thismachine =~ s/(\w+)(\..*)?/$1/;
	}
	return $thismachine;
    }
} # end of subroutine gethost

##################### replace_metacharacters ###########################
## Not all characters can be used 'as is' in a match. Some characters,
## called metacharacters, are reserved for use in regex notation.
## The metacharacters are
##    {}[]()^$.|*+?\
## A metacharacter can be matched by putting a backslash before it:
sub replace_metacharacters {
    my ($in,$extra) = @_;
    $in =~ s#\\#\\\\#g; # this must be first since it has \
    $in =~ s#\$#\\\$#g;
    $in =~ s#\(#\\(#g;
    $in =~ s#\)#\\)#g;
    $in =~ s#\{#\\{#g;
    $in =~ s#\}#\\}#g;
    $in =~ s#\[#\\[#g;
    $in =~ s#\]#\\]#g;
    $in =~ s#\^#\\^#g;
    $in =~ s#\.#\\.#g;
    $in =~ s#\|#\\|#g;
    $in =~ s#\*#\\*#g;
    $in =~ s#\+#\\+#g;
    $in =~ s#\?#\\?#g;
    if (defined($extra)) {
	$in =~ s#$extra#\\$extra#g;
    }
    return $in;
} # end of subroutine replace_metacharacters

####################### directory_status ##########################
## Check on status of directory and machine serving it using
## basic_computer_up and whoserve.
## Created 2/26/2002 by David Ovens.
## Required INPUT: $dir (use "" if you want it blank)
## Optional INPUTS: $outformat,$server,$senderrmail,$exitonfailure,
##                  $timeout,$maxfailcount,$mydebug
## $outformat = {"status","status,server"}
## status (DEFAULT) OUTPUT: 
## -1 = computer is undetermined, directory listing took too long or
##      directory does not yet exist.
##  0 = computer is down or unavailable
##  1 = computer is up, directory is ok
## status,server OUTPUT: status,machinename
sub directory_status {
    my($dir,$outformat,$server,$senderrmail,$exitonfailure,$timeout,$maxfailcount,$mydebug) = @_;
    my($i,$i4start,$i4now);
    $i4start = time;
    if ($server eq "" && $dir eq "") {
	&send_message("ERROR in $0 @saveargs\nyou have misused subroutine directory_status by not specifying a server or directory\n",1,$mail_list) if ($senderrmail);
	exit(-1) if ($exitonfailure);
	return -1;
    }
    if ($outformat ne "status,server") {
	$outformat = "status";
    }
    unless ($timeout) {
	$timeout = 15;
    }
    unless ($maxfailcount) {
	$maxfailcount = 3;
    }
    if ($server eq "") {
	for ($i=1; $i <= $maxfailcount; $i++) {
	    if ($dir =~ /^\/(tmp|var|root)/) {
		$server = &gethost();
	    } else {
		chomp($server = `/usr/local/wx/bin/whoserve $dir -limit $timeout -nomail`);
		if ($server !~ /MACHINE-NOT-FOUND-FOR/ && $server ne "") {
		    last;
		} else {
		    if ($i >= $maxfailcount) {
			$i4now = time;
			&send_message("ERROR in $0 @saveargs\nsubroutine directory_status could not find server for $dir in $timeout * $maxfailcount seconds\ni4start = $i4start\ni4now   = $i4now\n",1,$mail_list) if ($senderrmail);
			exit(-1) if ($exitonfailure);
			if ($outformat eq "status") {
			    return -1;
			} else {
			    return (-1,"MACHINE-NOT-FOUND-FOR-$dir");
			}
		    } else {
			sleep $timeout;
		    }
		}
	    }
	} # end of loop to find directory server with whoserve
    }
    for ($i=1; $i <= $maxfailcount; $i++) {
	if (&basic_computer_up($server,"",($timeout*$i),$mydebug)) {
	    if ($outformat eq "status") {
		return 1;
	    } else {
		return (1,$server);
	    }
	} elsif ($i == $maxfailcount) {
	    $i4now = time;
	    if ($senderrmail) {
		if ($dir ne "") {
		    &send_message("ERROR in $0 @saveargs\nsubroutine directory_status found server for $server for $dir but determined $server was down\nafter $maxfailcount attempts waiting from $timeout to $timeout * $maxfailcount seconds \ni4start = $i4start\ni4now  = $i4now\n",1,$mail_list);
		} else {
		    &send_message("ERROR in $0 @saveargs\n$server is down\nafter $maxfailcount attempts waiting from $timeout to $timeout * $maxfailcount seconds \ni4start = $i4start\ni4now  = $i4now\n",1,$mail_list);
		}
	    }
	    exit(1) if ($exitonfailure);
	    if ($outformat eq "status") {
		return 0;
	    } else {
		return (0,$server);
	    }
	} else {
	    $i4now = time;
	    ##
	    ## Added 4/7/2005 because &basic_computer_up can fail for
	    ## other reasons than the computer being down, i.e. too many
	    ## connections, and, in those cases, the repeat tests happen
	    ## immediately.  This sleep prevents that.
	    if ( ($i4now - $i4start) < ($timeout*$i) ) {
		sleep $timeout;
	    }
	}
	print "got past the basic_computer_up test in directory_status\n" if ($debug);
    }
    &send_message("ERROR in $0 @saveargs\nsubroutine directory_status HOW DID I GET HERE?\n",1,$mail_list);
    exit(1);
} # end of subroutine directory_status

########################### computer_up ################################
## Returns 1 if a computer is up, 0 if it is not.  This is safer than
## a directory check or a ping, because it requires that the computer
## be able to respond.  This does multiple checks, and so is more
## reliable than basic_computer_up (which this uses) the old version
## prior to 4/9/2002.
## Created by David Ovens.
## Completely rewritten by David Ovens 2/5/2002 to use sockets.
## Required INPUT: $computer
## Optional INPUTS: $pingcount, $timeout, $maxfailcount
## OUTPUT: 1 = computer is up, 0 = computer is down
sub computer_up {
    my ($computer,$pingcount,$timeout,$maxfailcount) = @_;
    ## NOTE: pingcount is obsolete, but should be retained for
    ##       legacy code.
    my ($i);
    if ($computer eq "") {
	print "ERROR in computer_up, you have not specified a computer\n";
	return 0;
    }
    unless ($timeout) {
	$timeout = 15;
    }
    unless ($maxfailcount) {
	$maxfailcount = 3;
    }
    for ($i=1; $i <= $maxfailcount; $i++) {
	if (&basic_computer_up($computer,"",($timeout*$i))) {
	    return 1;
	} elsif ($i == $maxfailcount) {
	    return 0;
	}
    }
} # end of subroutine computer_up
########################### computer_up ################################

###################### basic_computer_up ###############################
## Returns 1 if a computer is up, 0 if it is not.  This is safer than
## a directory check or a ping, because it requires that the computer
## be able to respond.
## Created 12/10/1999 by David Ovens.
## Completely rewritten by David Ovens 2/5/2002 to use sockets.
## Completely rewritten again by David Ovens 2/9/2003 to use
##   &basic_computer_up which is the only sure-fire way of making
##   sure that the socket gets closed properly and that code does not
##   Seg Fault.
## Renamed computer_up as basic_computer_up 4/9/2002 in order
##  to make the use of computer_up more reliable on failure messages.
## Required INPUT: $computer
## Optional INPUTS: $pingcount, $timeout
## OUTPUT: 1 = computer is up, 0 = computer is down
sub basic_computer_up {
    my ($computer,$pingcount,$timeout,$mydebug) = @_;
    ## NOTE: pingcount is obsolete, but should be retained for
    ##       legacy code.
    my ($host,$port,$name,$aliases,$type,$len,
	$proto,$thisaddr,$thataddr,$this,$that,$hostok,
	$sockaddr,$msg,$datalen,$a,$b,$c,$d,$ip,$username,
	@addrs);

    if ($mydebug eq "") {
	$mydebug = $debug;
    }
	
    ##
    ## It is important to have local $SIG{'ALRM'} = ... instead of
    ## putting $SIG{'ALRM'} into lines such as
    ##   my ($SIG{'ALRM'});
    ##   $SIG{'ALRM'} = sub {
    ## or
    ##   my $SIG{'ALRM'} = sub {
    ##
    use Sys::Hostname;
    $host = hostname;
    unless ($timeout) {
	$timeout = 15;
    }
    print "inside $$ checking on $computer with $timeout second timeout\n" if ($mydebug);
    use Socket;
    ##
    ## Get the port to check
#   $port = 512; # works for dept. machines
    $port = 22; # a secure shell, works for dept. and U.W. machines
    my ($time_left_on_previous_alarm);
    $sockaddr = 'S n a4 x8';
    ## $proto = "SSH-1.99-OpenSSH_3.4p1";
    $proto = getprotobyname('tcp');
    eval {
	local $SIG{ALRM} = sub { die "alarm\n" }; # NB: \n required
	## now die goes into $@
	$time_left_on_previous_alarm = alarm $timeout;
	$time_left_on_previous_alarm *= 1;

	($name,$aliases,$type,$len,@addrs) = gethostbyname($computer);
    
	$thataddr = $addrs[0];
	($a, $b, $c, $d) = unpack('C4', $thataddr);
	$ip = "$a.$b.$c.$d";
	print "ip for $name = $ip\n" if ($mydebug);
	if ($ip eq "...") {
	    die "no such machine: $computer\n";
	}
    
	($name,$aliases,$type,$len,@addrs) = gethostbyname($host);
	$thisaddr = $addrs[0];
    
	$this = pack($sockaddr,AF_INET,0,$thisaddr);
	$that = pack($sockaddr,AF_INET,$port,$thataddr);
    
	$hostok = 1;
	print "Connecting to $computer:$port with protocol $proto\n" if $mydebug;
	socket(SOCKET_HANDLE,PF_INET,SOCK_STREAM,$proto) || ($hostok = 0);
	if ($hostok == 0) {
	    close SOCKET_HANDLE;
	    die "socket: $!\n";
	}

	## the following bind makes Linux hang.....
#    bind(SOCKET_HANDLE,$this) || ($hostok = 0);
	if ($hostok == 0) {
	    close SOCKET_HANDLE;
	    die "socket: $!\n";
	}
	connect(SOCKET_HANDLE,$that) || ($hostok = 0); 
	if ($hostok == 0) { 
	    close( SOCKET_HANDLE );
	    die "connect: $!\nNow closing this failed attempt\n";
	}
	$username = getlogin || (getpwuid($<))[0] ;
	$msg = "$proto";
	$datalen = length $msg;

	syswrite SOCKET_HANDLE,$msg,$datalen;

	print "prior to attempting to actually connect to SOCKET_HANDLE got 
  these values:
    pid      = $$
    computer = $computer
    host     = $host
    name     = $name
    ip       = $ip
    timeout  = $timeout
    port     = $port
time_left_on_previous_alarm = $time_left_on_previous_alarm
    proto    = $proto
    hostok   = $hostok
    username = $username
    datalen  = $datalen
    \%id     = %id\n" if ($mydebug);
	$msg = <SOCKET_HANDLE>;

	print "got beyond msg, $msg\n" if ($mydebug);
	if ($msg !~ /SSH/) {
	    print "basic_computer_up ERROR. Bad message from host=$host: msg=:$msg:
  other values:
    computer = $computer
    host     = $host
    name     = $name
    ip       = $ip
    timeout  = $timeout
    port     = $port
time_left_on_previous_alarm = $time_left_on_previous_alarm
    proto    = $proto
    hostok   = $hostok
    username = $username
    datalen  = $datalen
    \%id     = %id

I will now attempt to exit instead of proceeding which has caused code
to crash in the past.
";
	    die "msg reading";
	}
	$msg = "%id\n";
#    $msg = "$username\n$username\n%id\n";
	$datalen = length $msg;

	syswrite SOCKET_HANDLE,$msg,$datalen;
	close(SOCKET_HANDLE);
    }; # all done with the eval
    my $now = time;
    if ($@ eq "alarm\n") {
#	die unless $@ eq "alarm\n";   # propagate unexpected errors
	# timed out
	print " TIMED OUT on $computer $timeout second check: reset alarm to time_left_on_previous_alarm = $time_left_on_previous_alarm and returning to $$ at i4time = $now, convert using time-increment $now i4 0 0 yd utc-local\n" if ($mydebug);
	alarm $time_left_on_previous_alarm;
	return 0;
    } elsif ($@) {
	print " FAILED FOR OTHER REASONS: $@ on $computer $timeout second check: reset alarm to time_left_on_previous_alarm = $time_left_on_previous_alarm and returning to $$ at i4time = $now, convert using time-increment $now i4 0 0 yd utc-local\n" if ($mydebug);
	alarm $time_left_on_previous_alarm;
	return 0;
    } else {
	# did not time out
	print "got all the way done on $computer $timeout second check, reset alarm to time_left_on_previous_alarm = $time_left_on_previous_alarm and returning to $$ at i4time = $now, convert using time-increment $now i4 0 0 yd utc-local\n" if ($mydebug);
	alarm $time_left_on_previous_alarm;
	return 1;
    }
    print "how did I get here? \n" if ($mydebug);
    return 0;
} # end of subroutine basic_computer_up
###################### basic_computer_up ###############################

############################## send_message ##############################
## For mailing messages to a list of users or to $LOGNAME.
## Required Inputs:
##    $msg:  The message.
##    $mail: Nonzero and nonempty values send the mail, others just print.
## Optional Inputs:
##    $list: List of recipients.  Use \@ in full addresses.
##    $subject: Subject line of mail.  Default is 'Error in $0'
##    $mailprog: Path to sendmail.
########################################################################
sub send_message {
    my ($msg,$mail,$list,$subject,$mailprog) = @_;
    my ($command);
    
    if ($mail) {
	if ($mailprog eq "") {
	    $mailprog = '/usr/lib/sendmail';
	    unless ($OS ne "") {
		my ($os);
		chomp($os = `uname -s`);
		$mailprog = "/usr/sbin/sendmail" if ($os eq "Linux");
	    }
	}
	if ($list     eq "") {
	    $list     = $ENV{LOGNAME};
	    if ($list eq "ldm") {
		$list = "stoves";
	    }
	}
	if ($subject  eq "") {
	    $command = $0;
	    $command =~ s#(.*/)?(.*)#$2#;
	    $subject  = "Error in $command";
	}

	open (MAIL, "|$mailprog \"$list\"") || die "Can't open $mailprog!\n";

	print MAIL "To: $list\n";
	print MAIL "Subject: $subject\n\n";
	print MAIL "$msg\n";
	close MAIL;
	sleep 3;
    } else {
	print "$msg\n";
    }
} # end of subroutine send_message
############################## send_message ##############################

##############################  abort   ##############################
## Used with $SIG{'ALRM'} = 'abort' to abort a script after a time limit.
sub abort {
#   global ($final_system_command)
    my($prog_id) = $$;
    chop($date = `date`);
    if (@savearg) {
	$message .= "ABORT time expired on $0 @savearg as of $date, so killed proc=$prog_id\n";
    } elsif (@saveargs) {
	$message .= "ABORT time expired on $0 @saveargs as of $date, so killed proc=$prog_id\n";
    } else {
	$message .= "ABORT time expired on $0 as of $date, so killed proc=$prog_id\n";
    }
    if (defined($mail_it)) {
	&send_message($message,$mail_it);
    } elsif (defined($mail_list)) {
	&send_message($message,1,$mail_list);
    } else {
	&send_message($message,1);
    }
    if ($final_system_command) { system "$final_system_command"; }
    $neg = -1 * $prog_id;
    if ($prog_id) {kill 9, $neg;}
    exit(1);
} # end of subroutine abort
############################## send_message ##############################

############################## count ##############################
## Count lines in a file.
## Believe it or not, it is faster to open the file and count all lines
## than it is to chop($count = `wc -l $file`);
sub count {
    my($in) = @_;
    my($count);
    eval {
	$count = 0;
	open(IN,"<$in") || die "Cannot open $in to count its lines";
	while(<IN>) {
	    $count++;
	}
	close IN;
    }; # all done with the eval
    if ($@) {
	print "$@";
	return -1;
    } else {
	return $count;
    }
    return $count;
}
############################## count ##############################

###################### find_remote_file #############################
## looks for a remote file from a remote machine, using ssh and an internal
## alarm to timeout if there's a problem
# Inputs:
#       $filename       --> full path to remote file
#       $remote_machine --> name of remote machine
#       $timeout        --> number of seconds to wait for the rsync (default = 120)
#       $remote_user    --> username on remote system (default = LOGNAME)
#
# Return value          --> 0 success, 1+ failure
#
sub find_remote_file {
    my($filename,$remote_machine,$timeout,$remote_user) = @_;
    my ($time_left_on_previous_alarm,$iret,$command);
    my $tmpfile = "/tmp/find_remote_file.stdout.$$";
    my (@more);
    if ($remote_user eq "" || $remote_user eq "default") {
	$remote_user = $ENV{LOGNAME};
    }
    if ($timeout eq "" || $timeout == 0) {
	$timeout = 120; # 2 minutes as a default
    }
    eval {
	local $SIG{ALRM} = sub { die "alarm\n" }; # NB: \n required
	## now die goes into $@
	$time_left_on_previous_alarm = alarm $timeout;
	$time_left_on_previous_alarm *= 1;
	$command = "ssh $remote_user\@$remote_machine ls -l \"$filename\" 1>$tmpfile 2>&1";
	print "$command\n";
	$iret = system "$command";
	open(IN,"<$tmpfile");
	@more = <IN>;
	close IN;
	unlink "$tmpfile";
    }; # all done with the eval
    my $now = time;
    if ($@ eq "alarm\n") {
#	die unless $@ eq "alarm\n";   # propagate unexpected errors
	# timed out
	print " TIMED OUT after $timeout seconds\nlocal file $localfilename may need to be deleted\nresetting alarm to time_left_on_previous_alarm = $time_left_on_previous_alarm\n";
	system "ls -l $localfilename";
	alarm $time_left_on_previous_alarm;
	return 1; # failure
    } elsif ($@) {
	print " FAILED FOR non-timeout REASONS: $@\n";
	alarm $time_left_on_previous_alarm;
	return 1; # failure
    } else {
	# did not time out
	alarm $time_left_on_previous_alarm;
	if ($iret != 0) {
	    print "@more";
	}
	return $iret; # success
    }
    print "how did I get here? \n" if ($mydebug);
    return 0;
} # end of subroutine find_remote_file
###################### find_remote_file #############################

####################### get_remote_file #############################
## retrieve a file using rsync from a remote machine, using an internal
## alarm to timeout if there's a problem
# Inputs:
#       $filename       --> full path to remote file
#       $remote_machine --> name of remote machine
#       $localfilename  --> full path to local file name
#       $timeout        --> number of seconds to wait for the rsync (default = 120)
#       $remote_user    --> username on remote system (default = LOGNAME)
#
# Return value          --> 0 success, 1+ failure
#
sub get_remote_file {
    my($filename,$remote_machine,$localfilename,$timeout,$remote_user) = @_;
    my ($time_left_on_previous_alarm,$iret,$command);
    my $tmpfile = "/tmp/get_remote_file.stdout.$$";
    my (@more);
    if ($remote_user eq "" || $remote_user eq "default") {
	$remote_user = $ENV{LOGNAME};
    }
    if ($timeout eq "" || $timeout == 0) {
	$timeout = 120; # 2 minutes as a default
    }
    eval {
	local $SIG{ALRM} = sub { die "alarm\n" }; # NB: \n required
	## now die goes into $@
	$time_left_on_previous_alarm = alarm $timeout;
	$time_left_on_previous_alarm *= 1;
	$command = "rsync -av $remote_user\@$remote_machine:$filename $localfilename 1>$tmpfile 2>&1";
	print "$command\n";
	$iret = system "$command";
	open(IN,"<$tmpfile");
	@more = <IN>;
	close IN;
	unlink "$tmpfile";
    }; # all done with the eval
    my $now = time;
    if ($@ eq "alarm\n") {
#	die unless $@ eq "alarm\n";   # propagate unexpected errors
	# timed out
	print " TIMED OUT after $timeout seconds\nlocal file $localfilename may need to be deleted\nresetting alarm to time_left_on_previous_alarm = $time_left_on_previous_alarm\n";
	system "ls -l $localfilename";
	alarm $time_left_on_previous_alarm;
	return 1; # failure
    } elsif ($@) {
	print " FAILED FOR non-timeout REASONS: $@\n";
	alarm $time_left_on_previous_alarm;
	return 1; # failure
    } else {
	# did not time out
	alarm $time_left_on_previous_alarm;
	if ($iret != 0) {
	    print "@more";
	}
	return $iret; # success
    }
    print "how did I get here? \n" if ($mydebug);
    return 0;
} # end of subroutine get_remote_file
############################ get_remote_file #############################

################################ ftp_file ################################
# This subroutine will loop for timelimit until a successful
#  anonymous ftp session has occurred.
# Inputs:
#         $ftpsite     --> where to get the file
#         $directory   --> where to look for the file
#         $file        --> what file to get
#         $outdir      --> where to put the file
#         $outfile     --> what to call the retrieved file
#         $timelimit   --> how long to keep trying
#         $sleeptime   --> how long to sleep between retries
# Inputs/Outputs:
#         *message     --> message variable
# Return value
#         $failure     --> 0 success, 1 not connected, 2 no such file/directory
#
sub ftp_file {
    ##
    ## NOTE: to use *message, need local, not my here!
    local ($ftpsite,$directory,$file,$outdir,$outfile
	   ,$timelimit,$sleeptime,*message) = @_;
    my ($savedir,$success,$failure,$date);

    $success = 0;
    chop($date = `date`);
    chop($savedir = `pwd`);
    chdir $outdir;
    
    print "retrieving $directory/$file from $ftpsite and putting in $outdir/$outfile at $date\n";

    open (FTP, ">/tmp/ftp$$");
    print FTP <<"END";
   open $ftpsite
   user anonymous $ENV{LOGNAME}\@atmos.washington.edu
   cd $directory
   binary
   get $file $outfile
   quit
END
    close (FTP);

    local $SIG{'ALRM'} = 'abort';  # IMPORTANT to be local
    alarm($timelimit);
    until ($success) {
	$failure = 0;
	open (FTP, "ftp -nv < /tmp/ftp$$ |");
	@ftp = <FTP>;
	close (FTP);
	foreach ( @ftp ) {
	    if (/Connection timed out/ || /Not connected/) {
		$failure = 1;
		chop($date = `date`);
		$message .= "$date, $ftpsite $directory/$file: $_\n";
		last;
	    } elsif ( /No such file or directory/ ) {
		$failure = 2;
		$message .= "ftp $ftpsite $directory/$file: wrong/missing directory\n";
		alarm(0);
		return $failure;
	    }
	}
	if ($failure) {
	    sleep $sleeptime;
	} else {
	    $success = 1;
	}
    }
    unlink ("/tmp/ftp$$");
    alarm(0);

    chop($date = `date`);
    print "finished retrieving $outfile at $date\n";
    chdir $savedir;
    return $failure;
} # end of subroutine ftp_file
################################ ftp_file ################################

################################ ftp_dir ################################
# This subroutine will loop for timelimit until a successful
#  anonymous ftp session has occurred.
# Inputs:
#         $ftpsite    --> site from which to obtain the listing
#         $directory  --> the directory to list (use commas for multiple)
#         $outfile    --> locally where to put the listing
#         $timelimit  --> how long to keep trying
#         $sleeptime  --> how long to sleep between retries
# Inputs/Outputs:
#         $message    --> message variable
sub ftp_dir {
    my ($ftpsite,$directory,$outfile
	   ,$timelimit,$sleeptime,$message) = @_;
    my ($success,$failure,$date,@dirs,$dir);
    $success = 0;
    chop($date = `date`);
    
    print "listing $directory from $ftpsite and putting in $outfile at $date\n";

    local $SIG{'ALRM'} = 'abort';  # we'll exit after $time_out seconds
    alarm($timelimit);
    until ($success) {
	$failure = 0;
	if (-e $outfile) {
	    unlink "$outfile";
	    if (-e $outfile) {
		print STDERR "ERROR: $0 cannot remove $outfile\n";
		exit 1;
	    }
	}
	if (index($directory,',') >= 0) {
	    @dirs = split(',',$directory);
	} else {
	    @dirs = $directory;
	}
	open (FTP, "|ftp -n > $outfile");
	print FTP "open $ftpsite
user anonymous $ENV{LOGNAME}\@atmos.washington.edu\n";
	foreach $dir (@dirs) {
	    print FTP "cd $dir\ndir\ncd /\n";
	}
	print FTP "quit\n";
	close FTP;
	open (TEST,"<$outfile");
	while (<TEST>) {
	    if (/Connection timed out/ || /Not connected/) {
		$failure = 1;
		chop($date = `date`);
		$message = $message . "$date, ftp: $_\n";
		last;
	    }
	}
	close TEST;
	if ($failure) {
	    sleep $sleeptime;
	} else {
	    $success = 1;
	}
    }
    alarm(0);

    chop($date = `date`);
    print "finished retrieving $directory in $outfile at $date\n";
    chdir $savedir;
}
################################ ftp_dir ################################

################################ nint ################################
## Returns nearest integer in a logical sense for positive and negative
## numbers.  Added 2/17/2005 after discovered that -12350 was becoming
## -12349 because of using int($num+0.5)!
sub nint {
    my ($nint) = @_;
    if ($nint >= 0) {
	return int($nint+0.5);
    } else {
	return int($nint-0.5);
    }
} # end of subroutine nint
################################ nint ################################

################################ max ################################
## Returns maximum value of a list or array.
sub max {
    my ($element,$max);
    foreach $element (@_) {
	if ($element ne "") {
	    $max = $element if ($max eq "");
	    $max = $element if ($element > $max);
	}
    }
    return $max;
} # end of subroutine max
################################ max ################################

################################ min ################################
## Returns minimum value of a list or array.
sub min {
    my ($element,$min);
    foreach $element (@_) {
	if ($element ne "") {
	    $min = $element if ($min eq "");
	    $min = $element if ($element < $min);
	}
    }
    return $min;
} # end of subroutine min
################################ min ################################

################################ sum ################################
## Returns sum of a list or array.
sub sum {
    my (@sum) = @_;
    my ($i,$sum,$missum,$numargs);
    $sum = $missum = 0;
    $numargs = $#sum + 1;
    while ($i = shift(@sum) ) {
	if ($i == $missvar) { $missum++; }
	if ($i eq " T" || $i eq "T" || $i eq "TR" || $i == $missvar) { $i = 0; }
	$sum = $sum + $i;
    }
    print "sum=$sum, missum=$missum, numargs=$numargs\n" if ($debug);
    if ($missum == $numargs) { return $missvar; }
    if ($sum == 0) { $sum = " T"; }
    return $sum;
} # end of subroutine sum
################################ sum ################################

############################### bynumber ################################
## Used with "sort bynumber" to sort numerically, rather than ASCII.
sub bynumber {
    $a <=> $b;
} # end of subroutine bynumber
############################### bynumber ################################

################################ std ################################
## Compute Standard Deviation.
##
## Global Input: $missvar
## Input:  @array
## Output: $standard_deviation,$mean.
##################################################################
sub std {
    my(@array) = @_;
    my($mean,$sum,$i);

    $mean = &mean(@array);
    if ($#array < 1) {
	print "Cannot use std with arrays smaller than 2 elements.\n" if ($debug);
	return $missvar,$mean;
    }
    $sum = $missum = 0;
    for ($i = 0; $i <= $#array; $i++) {
	if ($array[$i] == $missvar || $array[$i] eq "") {
	    $missum++;
	} else {
	    $sum += ($array[$i] - $mean)**2;
	}
    }
    if ($missum >= $#array) { return $missvar; }
    return sqrt($sum/$#array),$mean;
} # end of subroutine std
################################ std ################################

################################ mean ################################
## Compute mean of an array or list.
##
## Global Input: $missvar
## Output: mean (or missvar if no valid elements.
##
sub mean {
    my(@array) = @_;
    my ($i,$sum,$missum);
    $sum = $missum = 0;
    for ($i = 0; $i <= $#array; $i++) {
	if ($array[$i] == $missvar || $array[$i] eq "") {
	    $missum++;
	} else {
	    $sum += $array[$i];
	}
    } 
    if ($missum >= $#array+1) { return $missvar; }
    return $sum/($#array + 1);
} # end of subroutine mean
################################ mean ################################

##################### FileExists_BigEnough_NotChanging ##############
## Check on:
#            file existence
#            file size as compared to a given minimum size
#            file ceasing to grow in a specified time interval
## Once all three of these conditions are true, return 1, else 0.
## Additional bail out (11/6/2001):
##           set give_up_even_if_too_small to the number of times NOT seconds
##           that you will allow the file to be the same size before
##           bailing out, i.e. giving up and assuming that it just won't
##           get big enough.
sub FileExists_BigEnough_NotChanging {
    my ($file,$minsize,$changingTime,$fileSizeTimeout,$wait_till_exists,
	   $give_up_even_if_too_small,$quiet) = @_;
    my($iwaited,$ThisSize,$LastSize,$mail_it,$message,$num_same);
    
    if ($quiet == 1) {
	$mail_it = 0;
	$message = "";
    } else {
	$mail_it = 1;
	$message = "ERROR in subroutine FileExists_BigEnough_NotChanging in $0 for pid $$:\n";
    }
    $iwaited = $ThisSize = 0;
    $LastSize = -1;
    
  LOOP:
    if (-e "$file") {
	$ThisSize = (-s "$file");
	$num_same = 0;
	until ($ThisSize >= $minsize && $ThisSize == $LastSize) {
	    if ($ThisSize == $LastSize) {
		$num_same++;
	    } else {
		$num_same = 0;
	    }
	    if (defined($give_up_even_if_too_small) &&
		$num_same >= $give_up_even_if_too_small) {
		$message .= "We have waited too long ($num_same * $changingTime seconds)
(num_same = $num_same >= give_up_even_if_too_small = $give_up_even_if_too_small)
for $file to get big enough.\n";
		&send_message($message,$mail_it,$mail_list,$subject,$mailprog);
		return 0;
	    }
	    $LastSize = $ThisSize;
	    if ($iwaited < $fileSizeTimeout) {
		sleep $changingTime;
		$iwaited += $changingTime;
	    } else { # we have waited too long for this file to get big enough
		$message .= "We have waited too long ($iwaited >= $fileSizeTimeout) for $file ($LastSize) to get big enough.\n";
		&send_message($message,$mail_it,$mail_list,$subject,$mailprog);
		return 0;
	    }
	    $ThisSize = (-s "$file");
	} # size is big enough and it is no longer changing in size
	return 1;
    } else { # file does not exist yet
	if ($wait_till_exists) {
	    if ($iwaited < $fileSizeTimeout) {
		sleep $changingTime;
		$iwaited += $changingTime;
		goto LOOP;
	    } else { # we have waited too long for this file to get big enough
		$message .= "We have waited too long ($iwaited >= $fileSizeTimeout) for $file to exist.\n";
		&send_message($message,$mail_it,$mail_list,$subject,$mailprog);
		return 0;
	    }
	} else {
	    return 0;
	}
    }
} # end of subroutine FileExists_BigEnough_NotChanging
##################### FileExists_BigEnough_NotChanging ##############

############################### isrunning ########################
## returns 1 if process id $pid is still running
## returns 0 if $pid is not running
## returns -1 if we cannot tell because we cannot connect to the machine
## note: $firewall_machine is an optional argument
sub isrunning {
    my ($pid,$debug,$remote_machine,$firewall_machine,$user) = @_;
    ##
    ## Here begins the new, fast /proc/ version
    ##   UNFORTUNATELY, SOME MACHINES LEAVE DEFUNCT PROCESSES
    ##   AROUND....
    my ($ftest) = "-e";
    my $uid = "";
    my $dum;
    if ($user ne "") {
	$ftest = "-r";
	($dum,$dum,$uid) = getpwuid($>);
    }
    if ($firewall_machine ne "") {
	if ($remote_machine ne "") {
	    chomp($ans = `ssh $firewall_machine 'ssh -n $remote_machine \"if (-e /proc/$pid) echo 1\"'`);
	    if ($ans eq "1") {
		return 1;
	    } else {
		return 0;
	    }
	} else {
	    if (($user eq "" && -e "/proc/$pid") || ($user ne "" && -r "/proc/$pid/environ")) {
		##
		## added this to eliminate false return on defunct processes
		if (-e "/usr/local/wx/src/Proc-ProcessTable/lib") {
		    my ($proc,$ref);
		    eval "use Proc::ProcessTable";
		    if ($@) {
			##    print "\n\nfailure to load Proc::ProcessTable:\n $@...\n\n trying lib-x86_64 version\n";
			eval "use lib qw(/usr/local/wx/src/Proc-ProcessTable/lib)";
			eval "use lib qw(/usr/local/wx/src/Proc-ProcessTable/arch)";
			if ($@) {
##        print "trying x86_64 wx version...\n";
			    eval "use lib qw(/usr/local/wx/src/Proc-ProcessTable/lib-x86_64)";
			}
			eval "use Proc::ProcessTable";
			if ($@) {
##    print "last check trying x86_64 wx version...\n";
			    die "could not load ProcessTable ...";
			    exit;
			}
		    }
		    $ref = new Proc::ProcessTable;
		    foreach $proc (@{$ref->table}) {
			next unless ($proc->{pid} == $pid);
			if ($proc->{state} eq "defunct") {
			    return 0;
			} elsif ($proc->{cmndline} eq "") {
			    return 0;
			} elsif ($uid ne "" && $proc->{uid} ne $uid) {
			    return 0;
			} else {
			    return 1;
			}
		    }
		} else {
		    ##
		    ## cannot use ProcessTable stuff, assume it's running
		    &send_message("NOTE: you are trying to use the ProcessTable PERL module in the isrunning subroutine on a machine that does not have it installed.  you should ask support to copy it from an appropriate /usr/local/wx/src/Proc-ProcessTable distribution.  (location 1)",1);
		    return 1;
		}
	    } else {
		return 0;
	    }
	}
    } else {
	if ($remote_machine ne "") {
	    if ($user eq "") {
		chomp($ans = `ssh -n $remote_machine \"if (-e /proc/$pid) echo 1\"`);
		print "ssh -n $remote_machine \"if ($ftest /proc/$pid) echo 1\" gives $ans\n" if ($debug);
	    } else {
		chomp($ans = `ssh -n $remote_machine \"if (-r /proc/$pid/environ) echo 1\"`);
		print "ssh -n $remote_machine \"if ($ftest /proc/$pid/environ) echo 1\" gives $ans\n" if ($debug);
	    }
	    if ($ans eq "1") {
		return 1;
	    } else {
## removing the if debug clause on 7/25/2013 to help diagnose
## some network problems we've been having.  I don't think
## this will cause too much added logging and hassle.  D.Ovens
##		if ($debug) {
		    my $com = "ssh -n $remote_machine date";
		    print "$com gives: ";
		    $iret = system "$com";
		    print " with iret = $iret\n";
		    if ($iret != 0) {
			return -1;
		    } else {
			return 0;
		    }
##		}
		return 0;
	    }
	} else {
	    if (($user eq "" && -e "/proc/$pid") || ($user ne "" && -r "/proc/$pid")) {
		##
		## added this to eliminate false return on defunct processes
		if (-e "/usr/local/wx/src/Proc-ProcessTable/lib") {
		    my ($proc,$ref);
		    eval "use Proc::ProcessTable";
		    if ($@) {
			##    print "\n\nfailure to load Proc::ProcessTable:\n $@...\n\n trying lib-x86_64 version\n";
			eval "use lib qw(/usr/local/wx/src/Proc-ProcessTable/lib)";
			eval "use lib qw(/usr/local/wx/src/Proc-ProcessTable/arch)";
			if ($@) {
##        print "trying x86_64 wx version...\n";
			    eval "use lib qw(/usr/local/wx/src/Proc-ProcessTable/lib-x86_64)";
			}
			eval "use Proc::ProcessTable";
			if ($@) {
##    print "last check trying x86_64 wx version...\n";
			    die "could not load ProcessTable ...";
			    exit;
			}
		    }
		    $ref = new Proc::ProcessTable;
		    foreach $proc (@{$ref->table}) {
			next unless ($proc->{pid} == $pid);
			if ($proc->{state} eq "defunct") {
			    return 0;
			} elsif ($proc->{cmndline} eq "") {
			    return 0;
			} elsif ($uid ne "" && $proc->{uid} ne $uid) {
			    return 0;
			} else {
			    return 1;
			}
		    }
		} else {
		    ##
		    ## cannot use ProcessTable stuff, assume it's running
		    &send_message("NOTE: you are trying to use the ProcessTable PERL module in subroutine isrunning on a machine that does not have it installed.  you should ask support to copy it from an appropriate /usr/local/wx/src/Proc-ProcessTable distribution.  (location 2)",1);
		    return 1;
		}
	    } else {
		return 0;
	    }
	}
    }
##########################################################################
    ##
    ## Now follows the old version of this code with '##old' indents
##old    my ($OS,$newpid,$ppid,$job,$ps,$more) = "";
##old    if ($remote_machine ne "") {
##old	chomp($OS = `rsh -n $remote_machine uname -s`);
##old	$more = "rsh -n $remote_machine";
##old    } else {
##old	chomp($OS = `uname -s`);
##old	$more = "";
##old    }
##old    if ($OS eq "OSF1") {
##old	unless( open( PS, "$more /usr/ucb/ps alxww$pid |" ) ) {
##old	    print STDERR "Trouble running the ps command.";
##old	    exit 1;
##old	}
##old    } elsif ($OS eq "SunOS") {
##old	unless( open( PS, "$more /usr/ucb/ps alxww $pid |" ) ) {
##old	    print STDERR "Trouble running the ps command.";
##old	    exit 1;
##old	}
##old    } elsif ($OS eq "Linux") {
##old#	if (-e "/usr/ucb/ps") {
##old#	    $ps = "/usr/ucb/ps";
##old#	} else {
##old#	    $ps = "/bin/ps";
##old#	}
##old	$ps = "ps";
##old	unless( open( PS, "$more $ps lww $pid |" ) ) {
##old	    print STDERR "Trouble running the ps command.";
##old	    exit 1;
##old	}
##old    }
##old    @processes = <PS>; 
##old    close PS;
##old    foreach (@processes) {
##old	if ($OS eq "OSF1") {
##old	    if (/^\s*\d+\s+(\d+)\s+(\d+)\s.*:\d\d(\.\d\d)?\s*(\S.*)$/) {
##old		$newpid = $1;
##old		$ppid = $2;
##old		$job = $4;
##old	    }
##old	} elsif ($OS eq "SunOS" || $OS eq "Linux") {
##old	    if (/^\s*\d+\s+\d+\s+(\d+)\s+(\d+)\s.*:\d\d(\.\d\d)?\s*(\S.*)$/) {
##old		$newpid = $1;
##old		$ppid = $2;
##old		$job = $4;
##old	    }
##old	}
##old    } # end of foreach process
##old    print "process $pid child of $ppid running job $job\n" if ($debug);
##old    if ($job ne "" && $job !~ /<defunct>/) {
##old	return 1;
##old    } else {
##old	return 0;
##old    }
} # end of subroutine isrunning
############################### isrunning ########################

############################### source_gempak ########################
## will attempt to read /usr/local/gempak/Gemenviron and
## set all of the environment variables
## 
sub source_gempak {
    ##
    ## optional arguments: $file, local
    ## setting the local (by passing any non-zero second argument)
    ## means that this routine will fill local (non-environmental)
    ## variables.  So, for example, you'd be able to access
    ## MM5OUTDIR as $ENV{MM5OUTDIR} and $MM5OUTDIR
    ## dir_store as $ENV{dir_store} and $dir_store
    ## This is particularly useful for simple conversions of bash/csh
    ## scripts into PERL scripts.
    ##
    my($file,$local,@args) = @_;
    my $f1;
    my $f2;
    if ($file eq "") {
	$file = "/usr/local/gempak/Gemenviron";
    }
    if (! -e "$file") {
	return 1; # failure
    }
    $f1 = "/tmp/c1.$$";
    $f2 = "/tmp/c2.$$";
    ##
    ## take a quick look in the file to see if it is bash/bourne shell,
    ## otherwise, assume it is tcsh/csh.
    my $bourne = 0;
    if ($file =~ /.*\.(bash|sh)$/) {
	$bourne = 1;
    } elsif ($file =~ /.*\.(csh|tcsh)$/) {
	$bourne = 0;
    } else {
	unless (open(IN,"<$file")) {
	    print "ERROR: cannot open $file for reading\n";
	    return -1;
	}
	my $count = 0;
	while (<IN>) {
	    $count++;
	    last if ($count > 11);
	    if ($count == 1 && (/bin\/bash/ || /bin\/sh/)) {
		$bourne = 1;
		last;
	    } elsif (/\S=\w/) {
		## no spaces in variable assignments
		$bourne = 1;
	    }
	}
	close IN;
    }
    my %f1;
    my %f2;
    if ($bourne == 0) {
	system "csh -fc \"printenv >! $f1; source $file @args; printenv >! $f2\" ";
	open(IN,"<$f1");
	while (<IN>) {
	    next if (/PIPESTATUS/); # a strange variable that causes errors
	    if (/^(\S+)=(.*)$/) {
		$var = $1;
		$val = $2;
		$f1{$var} = $val;
	    }
	}
	close IN;
    } else {
	system "set > $f1; source $file @args; set > $f2";
	open(IN,"<$f1");
	while (<IN>) {
	    next if (/PIPESTATUS/); # a strange variable that causes errors
	    if (/^(\S+)=(.*)$/) {
		$var = $1;
		$val = $2;
		$f1{$var} = $val;
	    }
	}
	close IN;
    }
    if ($local) {
	unlink "$f1";
	##
	## now setting allowing $ENV{dir_store} or $ENV{MM5OUTDIR} to
	## be accessed from PERL as $dir_store or $MM5OUTDIR.
	## create a simple *.pl file and use require.
	open(OUT,">$f1");
    }
    open(IN,"<$f2");
    while (<IN>) {
	next if (/PIPESTATUS/); # a strange variable that causes errors
	if (/^(\S+)=(.*)$/) {
	    $var = $1;
	    $val = $2;
	    $f2{$var} = $val;
	    if ($f2{$var} ne $f1{$var}) {
		$ENV{$var} = $val;
		if ($local) {
		    ## skip anything that does not start with a letter or number
		    
		    print OUT sprintf ("\$%s = \"%s\";\n",$var,$val);
		}
	    }
	}
    }
    close IN;
    if ($local) {
	print OUT "1;\n";
	close OUT;
	require "$f1";
    }
    unlink "$f1","$f2";
    return 0; # success
} # end of subroutine source_gempak

############################### source_gempak ########################
## will source a file in CSH and fill %ENV in PERL
## 
sub source_file_fill_env {
    ##
    ## uses source_gempak
    my ($file,$local,@args) = @_;
    my $iret;
    $iret = &source_gempak($file,$local,@args);
    return $iret;
} # end of subroutine source_file_fill_env
############################### source_gempak ########################

############################### gdfind ##################################
## tells us if a parameter, level, coordinate, forecast hour is found
## in a GEMPAK file
## SEE ALSO gdfind_array which is faster for a whole array of data
## Author: David Ovens (ovens@atmos.washington.edu) 2/28/2006
sub gdfind {
    ##
    ## INPUTS: $file,$gfunc,$glevel,$gvcord,$fhr
    ## OPTIONAL GLOBAL INPUTS: $debug, $0, @ARGV
    ## OUTPUT: 1 if found, 0 if not found, -1 if cannot run gdlist
    ## SEE ALSO gdfind_array which is faster for a whole array of data
    my ($file,$gfunc,$glevel,$gvcord,$fhr) = @_;
    my $fffhr;
    my $dumfile;
    my $found;

    $fffhr = sprintf "%3.3d",$fhr;
    $dumfile = "/tmp/gdlist$$";
    $found = 0;
    $gfunc  =~ tr/a-z/A-Z/;
    $gvcord =~ tr/a-z/A-Z/;
    
    unless (open(GDINFO,"|gdinfo > /dev/null")) {
	print "ERROR in $0 @ARGV: cannot run gdinfo in sub gdfind @_\n";
	return -1;
    }
    print GDINFO " GDFILE   = $file
 LSTALL   = yes
 OUTPUT   = f/$dumfile
 GDATTIM  = F$fffhr
 GLEVEL   = $glevel
 GVCORD   = $gvcord
 GFUNC    = $gfunc
 run

exit\n";
    close GDINFO;
    print " GDFILE   = $file
 LSTALL   = yes
 OUTPUT   = f/$dumfile
 GDATTIM  = F$fffhr
 GLEVEL   = $glevel
 GVCORD   = $gvcord
 GFUNC    = $gfunc
 run

exit\n" if ($debug);
    unless (open(GDINFO,"<$dumfile")) {
	print "ERROR in $0 @ARGV: cannot open $dumfile in gdfind @_\n";
	return -1;
    }
    while (<GDINFO>) {
	if (/F$fffhr\s+$glevel\s+$gvcord\s+$gfunc/) {
	    $found = 1;
	    last;
	}
    }
    close GDINFO;
    unlink "$dumfile";
    return $found
} # end of subroutine gdfind
############################### gdfind ##################################

########################### gdfind_array ################################
## tells us if a parameter, level, coordinate, forecast hour is found
## in a GEMPAK file
## Author: David Ovens (ovens@atmos.washington.edu) 2/28/2006
sub gdfind_array {
    ##
    ## INPUTS: $file,@array_of($gfunc,$glevel,$gvcord,$fhr)
    ## OPTIONAL GLOBAL INPUTS: $debug, $0, @ARGV
    ## OUTPUT: 1 if found, 0 if not found, -1 if cannot run gdlist
    my ($file,@array) = @_;
    my $gfunc;
    my $glevel;
    my $gvcord;
    my $fhr;
    my $fffhr;
    my $dumfile;
    my $found;
    my $rec_term;
    my $gdinfo;
##  GLOBAL: $gdfind_array_missed
    $dumfile = "/tmp/gdlist$$";
    $found = 1;
    
    unless (open(GDINFO,"|gdinfo > /dev/null")) {
	print "ERROR in $0 @ARGV: cannot run gdinfo in sub gdfind @_\n";
	return -1;
    }
    print GDINFO " GDFILE   = $file
 LSTALL   = yes
 OUTPUT   = f/$dumfile
 GDATTIM  = all
 GLEVEL   = all
 GVCORD   = all
 GFUNC    = all
 run

exit\n";
    close GDINFO;
    print " GDFILE   = $file
 LSTALL   = yes
 OUTPUT   = f/$dumfile
 GDATTIM  = all
 GLEVEL   = all
 GVCORD   = all
 GFUNC    = all
 run

exit\n" if ($debug);
    unless (open(GDINFO,">>$dumfile")) {
	print "ERROR in $0 @ARGV: cannot open $dumfile in gdfind @_\n";
	return -1;
    }
    print GDINFO "\nZCZCZC\n";
    close GDINFO;
    $rec_term = $/;
    $/ = "ZCZCZC";
    unless (open(GDINFO,"<$dumfile")) {
	print "ERROR in $0 @ARGV: cannot open $dumfile in gdfind @_\n";
	return -1;
    }
    $gdinfo = <GDINFO>; # slurp it all in as one line
    close GDINFO;
    unlink "$dumfile";
    foreach (@array) {
	($gfunc,$glevel,$gvcord,$fhr) = split(',',$_);
	$fffhr = sprintf "%3.3d",$fhr;
	$gfunc  =~ tr/a-z/A-Z/;
	$gvcord =~ tr/a-z/A-Z/;
	if ($gdinfo !~ /F$fffhr\s+$glevel\s+$gvcord\s+$gfunc/) {
	    print " gdfind_array:  missed $gfunc $glevel $gvcord at hour $fhr\n";
	    $gdfind_array_missed = "$gfunc,$glevel,$gvcord,$fhr";
	    $found = 0;
	    last;
	}
    }
    $/ = $rec_term;
    return $found
} # end of subroutine gdfind_array
########################### gdfind_array ################################

########################### getnodes ################################
## tells us the number of nodes in the named cluster
## INPUT: machine name: ens, sage1, n1, n2, tahoma, rainier, etc.
## OUTPUT: array of machine names in the cluster
## GLOBAL OUTPUT: getnodes_slots (number of slots per node),
##                getnodes_cluster (base name of the cluster machines)
##                getnodes_cluster_server (name of the cluster, aka the main server)
##                %slotsbynode (number of slots for each node)
## Usefulness:  Allows users to update only this file and use the
## exact same code to find all of the machines in the cluster.
## Created:  David Ovens 5/7/2007
sub getnodes {
    my ($machine) = @_;
    my @machines = "";
    my $i;
    our ($getnodes_slots,$getnodes_cluster,$getnodes_cluster_server,%slotsbynode) = "";
    if ($machine eq "") {
	$machine = &gethost;
    }
    if ($machine =~ /^ens/) {
	$getnodes_slots = 4;
	$getnodes_cluster = "ens"; # base name of each machine
	$getnodes_cluster_server = "ens"; # base name of the cluster's server or main machine
	@machines = qw( ens1 ens2 ens3 ens4);
	foreach $i (@machines) {
	    $slotsbynode{$i} = $getnodes_slots;
	}
	return @machines;
    } elsif ($machine eq "sage") {
	## new machine returns
##	$getnodes_slots = 8;
##	$getnodes_cluster = "sage";
##	return ( "$machine" );
	## old machine needs to use the other nodes, it
	## cannot run all of the code that the other nodes can
	## updated 4/14/2008
	## added n15 n16 n17 on 7/1/2011, and again 8/16/2011
        ## added n18 n19 n20 n21 on 20/17/2011
	$getnodes_slots = 20;
	$getnodes_cluster = "n"; # base name of each machine
	$getnodes_cluster_server = "sage"; # base name of the cluster's server or main machine
#	@machines = qw( n1 n2 n3 n4 n5 n6 n7 n8 n15 n16 n17 n18 n19 n20 n21);
#       added n59-n62 3/9/2016	
#       added n21-n28 8/29/2018
	@machines = ( "n51", "n52", "n53", "n54", "n55", "n56", "n57", "n58", "n59", "n60", "n61", "n62", "n63", "n64", "n65", "n66", "n67", "n68", "n69", "n70" );
	foreach $i (@machines) {
	    $slotsbynode{$i} = $getnodes_slots;
	}
	$slotsbynode{n63} = $slotsbynode{n64} = $slotsbynode{n65} = $slotsbynode{n66} = $slotsbynode{n67} = $slotsbynode{n68} = $slotsbynode{n69} = $slotsbynode{n70} = 24;
	$slotsbynode{n21} = $slotsbynode{n22} = $slotsbynode{n23} = $slotsbynode{n24} = $slotsbynode{n25} = $slotsbynode{n26} = $slotsbynode{n27} = $slotsbynode{n28} = 24;
	return @machines;
    } elsif ($machine =~ /^(sage|n)\d+/) {
	$getnodes_slots = 20;
	$getnodes_cluster = "n"; # base name of each machine
	$getnodes_cluster_server = "sage"; # base name of the cluster's server or main machine
#	@machines = qw( n1 n2 n3 n4 n5 n6 n7 n8 n15 n16 n17 n18 n19 n20 n21);
#       added n59-n62 3/9/2016	
	@machines = ( "n51", "n52", "n53", "n54", "n55", "n56", "n57", "n58", "n59", "n60", "n61", "n62", "n63", "n64", "n65", "n66", "n67", "n68", "n69", "n70" );
	foreach $i (@machines) {
	    $slotsbynode{$i} = $getnodes_slots;

	}
	$slotsbynode{n63} = $slotsbynode{n64} = $slotsbynode{n65} = $slotsbynode{n66} = $slotsbynode{n67} = $slotsbynode{n68} = $slotsbynode{n69} = $slotsbynode{n70} = 24;
	$slotsbynode{n21} = $slotsbynode{n22} = $slotsbynode{n23} = $slotsbynode{n24} = $slotsbynode{n25} = $slotsbynode{n26} = $slotsbynode{n27} = $slotsbynode{n28} = 24;
	return @machines;
    } elsif ($machine =~ /^aorr/) {
	$getnodes_slots = 12;
	$getnodes_cluster = "a"; # base name of each machine
	$getnodes_cluster_server = "aorr"; # base name of the cluster's server or main machine
#	@machines = qw( a109 a110 a111 a112 a113 a114 a115 a116 a117 a118 a119 a120 a121 a122);
	@machines = qw( a109 a110 a111 a112 a113 a114 a115 a116 a117 a118 a119 a120);
	foreach $i (@machines) {
	    $slotsbynode{$i} = $getnodes_slots;
	}
	return @machines;
    } elsif ($machine =~ /^aor/) {
	$getnodes_slots = 8;
	$getnodes_cluster = "a"; # base name of each machine
	$getnodes_cluster_server = "aor"; # base name of the cluster's server or main machine
	@machines = qw( a1 a2 a3 a4 a5 a6 a7 a8);
	foreach $i (@machines) {
	    $slotsbynode{$i} = $getnodes_slots;
	}
	return @machines;
    } elsif ($machine =~ /^(sage|n)$/) {
	$getnodes_slots = 20;
	$getnodes_cluster = "n"; # base name of each machine
	$getnodes_cluster_server = "sage"; # base name of the cluster's server or main machine
#	@machines = qw( n1 n2 n3 n4 n5 n6 n7 n8 n15 n16 n17 n18 n19 n20 n21);
#       added n59-n62 3/9/2016	
	@machines = ( "n51", "n52", "n53", "n54", "n55", "n56", "n57", "n58", "n59", "n60", "n61", "n62", "n63", "n64", "n65", "n66", "n67", "n68", "n69", "n70" );
	foreach $i (@machines) {
	    $slotsbynode{$i} = $getnodes_slots;
	}
	$slotsbynode{n63} = $slotsbynode{n64} = $slotsbynode{n65} = $slotsbynode{n66} = $slotsbynode{n67} = $slotsbynode{n68} = $slotsbynode{n69} = $slotsbynode{n70} = 24;
	$slotsbynode{n21} = $slotsbynode{n22} = $slotsbynode{n23} = $slotsbynode{n24} = $slotsbynode{n25} = $slotsbynode{n26} = $slotsbynode{n27} = $slotsbynode{n28} = 24;
	return @machines;
    } elsif ($machine =~ /^(tahoma|rainier)/) {
	$getnodes_slots = 4;
	$getnodes_cluster = "$machine"; # base name of each machine
	$getnodes_cluster_server = "$machine"; # base name of the cluster's server or main machine
	%slotsbynode = ($machine => $getnodes_slots);
	return ( "$machine" );
    } elsif ($machine =~ /^(enkf)/) {
	$getnodes_slots = 8;
	$getnodes_cluster = "enkf"; # base name of each machine
	$getnodes_cluster_server = "$enkf"; # base name of the cluster's server or main machine
	return qw( enkf1 enkf2 enkf3);
    } elsif (-e "/proc/cpuinfo") {
	##
	## read /proc/cpuinfo and look for unique pairs of
	## physical id and core id
	my %junk;
	my $lastphysicalid;
	my @in;
	my ($key1,$key2) = "";
	open(IN,"cat /proc/cpuinfo |");
	@in = <IN>;
	close IN;
	foreach (@in) {
	    if (/physical id.*: (\d+)/) {
		$lastphysicalid = $1;
	    } elsif (/core id.*: (\d+)/) {
		$junk{$lastphysicalid}{$1} = 1;
	    }
	}
	$getnodes_slots = 0;
	foreach $key1 (keys(%junk)) {
	    foreach $key2 (keys(%{$junk{$key1}})) {
		$getnodes_slots++;
	    }
	}
	$getnodes_cluster = "$machine"; # base name of each machine
	$getnodes_cluster_server = "$machine"; # base name of the cluster's server or main machine
	%slotsbynode = ($machine => $getnodes_slots);
	return ( "$machine" );
    } else {
	$getnodes_slots = 1;
	%slotsbynode = ($machine => $getnodes_slots);
	return ( "$machine" );
    }
} # end of subroutine getnodes
########################### getnodes ################################

########################### dirspd2uv ################################
## Added 4/12/2019 by David Ovens
## calculate U and V components of wind given direction in degrees
## and speed.  U and V will be in the same units as speed
sub dirspd2uv {
    ##
    ## compute the U and V components of wind speed given
    ## direction and speed
    my($dir,$spd) = @_;
    my $u,$v;
    unless ($degrees2radians) {
	unless ($pi) {
	    my $x = sqrt(2.) / 2.;
	    $pi = 4. * atan2($x,$x);
	}
	$degrees2radians = $pi / 180.;
    }
    if ($spd == 0.) {
	return (0.,0.);
    } else {
	$u = -sin($dir*$degrees2radians) * $spd;
	$v = -cos($dir*$degrees2radians) * $spd;
	$u = 0 if (abs($u) <= 1e-15);
	$v = 0 if (abs($v) <= 1e-15);
	return ($u,$v);
    }
} # end of subroutine dirspd2uv
########################### dirspd2uv ################################

########################### uv2dirspd ################################
## Added 4/12/2019 by David Ovens
## Given U and V components of the wind, calculate direction in degrees
## and speed in the same units as U and V.
sub uv2dirspd {
    my($u,$v) = @_;
    my $dir,$spd;
    return (360.,0.) if ($u == 0 && $v == 0);
    unless ($radians2degrees) {
	unless ($pi) {
	    my $x = sqrt(2.) / 2.;
	    $pi = 4. * atan2($x,$x);
	}
	$radians2degrees = 180. / $pi;
    }
    $dir = atan2(-$u,-$v)*$radians2degrees;
    $spd = sqrt($u*$u + $v*$v);
    $dir += 360. if ($dir <= 0.);
    return($dir,$spd);
} # end of subroutine uv2dirspd
########################### uv2dirspd ################################

########################### dirspd2uv ################################
## Added 10/28/2020 by David Ovens
## calculate relative humidity given temperature and dewpoint in Celsius
sub relh {
    ##
    ## RELH  =  VAPR / VAPS * 100
    ## VAPR = vapor pressure
    ## VAPS = saturation vapor pressure
    my($tempC,$dewC) = @_;
    my $e,$esat;
    if ($tempC < -240 || $dewC < -240) {
	return -999;
    } else {
	# find the vapor pressure
	$e = 6.112 * exp ( ( 17.67 * $dewC ) / ( $dewC + 243.5 ) );
	# find the saturation vapor pressure
	$esat = 6.112 * exp ( ( 17.67 * $tempC ) / ( $tempC + 243.5 ) );
	return 100. * ($e/$esat);
    }
} # end of subroutine relh
########################### relh ################################

########################### FtoC ################################
## Added 10/28/2020 by David Ovens
## convert degrees Fahrenheit to degrees Celsius
sub FtoC {
    ##
    my($tempF) = @_;
    return ($tempF - 32.) / 1.8;
} # end of subroutine FtoC
########################### FtoC ################################

########################### CtoF ################################
## Added 10/28/2020 by David Ovens
## convert degrees Fahrenheit to degrees Celsius
sub CtoF {
    ##
    my($tempC) = @_;
    return ($tempC * 1.8) + 32.;
} # end of subroutine CtoF
########################### CtoF ################################


1;
## end from my local-lib.pl
########################################################################
